! ------------------------------------------------------------------------
!>  \brief Module defining Banded Symmetric Matrix data type
! ------------------------------------------------------------------------
MODULE band_sym_matrix_def
USE log_message_control   !> Printing log/error messages
USE log_messages          !> Log/error codes
USE vector_def            !> Vector data type
USE dense_matrix_def      !> Dense Matrix data type

IMPLICIT NONE
PRIVATE



! ************************************************************************
! EXPORTS
! ************************************************************************


!> Exported data types
PUBLIC :: bandSymMatrixT

!> Exported interfaces
PUBLIC :: bsm_init, bsm_clean, &
          bsm_numRows, bsm_halfBW, &
          bsm_get, bsm_set, &
          bsm_setDecomp, bsm_isDecomposed, &
          bsm_add, OPERATOR (+), bsm_mappedAdd, &
          bsm_scalMul, OPERATOR (*), bsm_vecMul



! ************************************************************************
! LOCAL CONSTANTS
! ************************************************************************

!> Sender code for Banded SYmmetric MATrix module
INTEGER, PARAMETER :: sdr = BSYMAT



! ************************************************************************
! DATA TYPES
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Banded Symmetric Matrix ADT structure
!!
!!  \param dat            Array to store packed (banded symmetric) data
!!  \param decomp         Array to store packed Cholesky decomposition
!!  \param is_decomposed  Switch indicating whether decomposition is present
! ------------------------------------------------------------------------
TYPE bandSymMatrixT
  DOUBLE PRECISION, ALLOCATABLE :: dat(:,:)
  DOUBLE PRECISION, ALLOCATABLE :: decomp(:,:)
  LOGICAL :: is_decomposed
END TYPE bandSymMatrixT



! ************************************************************************
! INTERFACES
! ************************************************************************


!>  \brief Interface to constructor
INTERFACE bsm_init
  MODULE PROCEDURE bsm_init_
  MODULE PROCEDURE bsm_init_exc_
END INTERFACE bsm_init

!>  \brief Interface to destructor
INTERFACE bsm_clean
  MODULE PROCEDURE bsm_clean_
END INTERFACE bsm_clean

!>  \brief Interface to number of rows
INTERFACE bsm_numRows
  MODULE PROCEDURE bsm_num_rows_
END INTERFACE bsm_numRows

!>  \brief Interface to half bandwidth
INTERFACE bsm_halfBW
  MODULE PROCEDURE bsm_half_bw_
END INTERFACE bsm_halfBW

!>  \brief Interface to getter for individual entries in matrix
INTERFACE bsm_get
  MODULE PROCEDURE bsm_get_
  MODULE PROCEDURE bsm_get_exc_
END INTERFACE bsm_get

!>  \brief Interface to setter for individual entries in matrix
INTERFACE bsm_set
  MODULE PROCEDURE bsm_set_
  MODULE PROCEDURE bsm_set_exc_
END INTERFACE bsm_set

!>  \brief Interface to setter for Cholesky decomposition
INTERFACE bsm_setDecomp
  MODULE PROCEDURE bsm_set_decomp_
  MODULE PROCEDURE bsm_set_decomp_exc_
END INTERFACE bsm_setDecomp

!>  \brief Interface to check if decomposition exists
INTERFACE bsm_isDecomposed
  MODULE PROCEDURE bsm_is_decomposed_
END INTERFACE bsm_isDecomposed

!>  \brief Interface for addition of two bandSymMatrixT
INTERFACE bsm_add
  MODULE PROCEDURE bsm_add_
  MODULE PROCEDURE bsm_add_exc_
END INTERFACE bsm_add

!>  \brief Operator overload for addition
INTERFACE OPERATOR (+)
  MODULE PROCEDURE bsm_add_
END INTERFACE

!>  \brief Interface for adding a matrixT to a bandSymMatrixT in place with mapping indices
INTERFACE bsm_mappedAdd
  MODULE PROCEDURE bsm_mapped_add_
  MODULE PROCEDURE bsm_mapped_add_exc_
END INTERFACE bsm_mappedAdd

!>  \brief Interface for scalar multiplication
INTERFACE bsm_scalMul
  MODULE PROCEDURE bsm_scal_mul_mat_scal_
END INTERFACE bsm_scalMul

!>  \brief Interface for matrix-vector multiplication
INTERFACE bsm_vecMul
  MODULE PROCEDURE bsm_vec_mul_
  MODULE PROCEDURE bsm_vec_mul_exc_
END INTERFACE bsm_vecMul

!>  \brief Operator overload for multiplication (includes scalar and vector multiplication)
INTERFACE OPERATOR (*)
  MODULE PROCEDURE bsm_scal_mul_mat_scal_
  MODULE PROCEDURE bsm_scal_mul_scal_mat_
  MODULE PROCEDURE bsm_vec_mul_
END INTERFACE



CONTAINS



! ************************************************************************
! ACCESS PROGRAMS
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Constructor for bandSymMatrixT (non-exception checking)
!!
!!  \param self Reference to the banded symmetric matrix to be initialized
!!  \param hbw  Half bandwidth of the matrix
!!  \param n    Number of rows/columns in the matrix
!!
!!  This routine initializes the bandSymMatrixT object referenced by self.
!!  The data structures contained in self are allocated and initial values
!!  are set to zero.
! ------------------------------------------------------------------------
SUBROUTINE bsm_init_ (self, hbw,n)
  TYPE(bandSymMatrixT), INTENT(INOUT) :: self
  INTEGER, INTENT(IN) :: hbw,n
  INTEGER :: i,j    !> loop variables

  !> only reallocate if new dimensions do not match existing dimensions
  IF (bsm_halfBW(self).NE.hbw .OR. bsm_numRows(self).NE.n) THEN

    !> reset object if it was previously initialized
    CALL bsm_clean(self)

    !> allocate new data structures
    ALLOCATE(self%dat(hbw,n))
    ALLOCATE(self%decomp(hbw,n))

  END IF

  !> ensure data is initialized to zero
  DO j = 1,n
    DO i = 1,hbw
      self%dat(i,j) = 0.d0
      self%decomp(i,j) = 0.d0
    END DO
  END DO

  !> reset boolean for decomposition
  self%is_decomposed = .FALSE.

END SUBROUTINE bsm_init_


! ------------------------------------------------------------------------
!>  \brief Constructor for bandSymMatrixT (exception checking)
!!
!!  \param self Reference to the banded symmetric matrix to be initialized
!!  \param hbw  Half bandwidth of the matrix
!!  \param n    Number of rows/columns in the matrix
!!  \param exc  Error code
!!
!!  This routine initializes the bandSymMatrixT object referenced by self.
!!  The data structures contained in self are allocated and initial values
!!  are set to zero. If the specified dimensions are invalid, a SZE
!!  exception is returned. If allocation of data structures fails, an
!!  ALLOC exception is returned.
! ------------------------------------------------------------------------
SUBROUTINE bsm_init_exc_ (self, hbw,n, exc)
  TYPE(bandSymMatrixT), INTENT(INOUT) :: self
  INTEGER, INTENT(IN) :: hbw,n
  INTEGER, INTENT(OUT) :: exc
  INTEGER :: e      !> allocation info code
  INTEGER :: i,j    !> loop variables

  !> if dimensions are invalid, raise SZE exception
  IF (hbw.LE.0 .OR. n.LE.0 .OR. hbw.GT.n) THEN
    exc=SZE
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> only reallocate if new dimensions do not match existing dimensions
  IF (bsm_halfBW(self).NE.hbw .OR. bsm_numRows(self).NE.n) THEN

    !> reset object if it was previously initialized
    CALL bsm_clean(self)

    !> allocate new data structure for matrix storage
    ALLOCATE(self%dat(hbw,n), STAT=e)

    !> if error code is returned, raise ALLOC exception
    IF (e.NE.0) THEN
      exc=ALLOC
      CALL log_printLogMsg(exc,sdr)
      CALL bsm_clean(self)
      RETURN
    ELSE
      exc=OK
    END IF

    !> allocate new data structure for decomposition storage
    ALLOCATE(self%decomp(hbw,n), STAT=e)

    !> if error code is returned, raise ALLOC exception
    IF (e.NE.0) THEN
      exc=ALLOC
      CALL log_printLogMsg(exc,sdr)
      CALL bsm_clean(self)
      RETURN
    ELSE
      exc=OK
    END IF

  END IF

  !> ensure data is initialized to zero
  DO j = 1,n
    DO i = 1,hbw
      self%dat(i,j) = 0.d0
      self%decomp(i,j) = 0.d0
    END DO
  END DO

  !> reset boolean for decomposition
  self%is_decomposed = .FALSE.

END SUBROUTINE bsm_init_exc_


! ------------------------------------------------------------------------
!>  \brief Destructor for bandSymMatrixT
!!
!!  \param self Reference to the banded symmetric matrix to be destroyed
!!
!!  This routine deallocates existing data structures and resets the
!!  switch indicating decomposition.
! ------------------------------------------------------------------------
SUBROUTINE bsm_clean_ (self)
  TYPE(bandSymMatrixT), INTENT(INOUT) :: self

  !> deallocate data structures
  IF (ALLOCATED(self%dat)) DEALLOCATE(self%dat)
  IF (ALLOCATED(self%decomp)) DEALLOCATE(self%decomp)

  !> reset boolean for decomposition
  self%is_decomposed = .FALSE.

END SUBROUTINE bsm_clean_


! ------------------------------------------------------------------------
!>  \brief Getter for number of rows
!!
!!  \param  self  Reference to the banded symmetric matrix object
!!  \return n     Number of rows in the matrix
!!
!!  This routine determines the number of rows allocated to the matrix
!!  object. If the matrix is not initialized it returns 0.
! ------------------------------------------------------------------------
FUNCTION bsm_num_rows_ (self) RESULT(n)
  TYPE(bandSymMatrixT), INTENT(IN) :: self
  INTEGER :: n

  !> if data is initialized, return number of rows
  IF (ALLOCATED(self%dat)) THEN
    n = SIZE(self%dat,2)
  ELSE
    n = 0   !> if not initialized, there are no rows
  END IF

END FUNCTION bsm_num_rows_


! ------------------------------------------------------------------------
!>  \brief Getter for half bandwidth
!!
!!  \param  self  Reference to the banded symmetric matrix object
!!  \return hbw   Half bandwidth of the matrix
!!
!!  This routine determines the half bandwidth allocated to the matrix
!!  object. If the matrix is not initialized it returns 0.
! ------------------------------------------------------------------------
FUNCTION bsm_half_bw_ (self) RESULT(hbw)
  TYPE(bandSymMatrixT), INTENT(IN) :: self
  INTEGER :: hbw

  !> if data is initialized, return half bandwidth
  IF (ALLOCATED(self%dat)) THEN
    hbw = SIZE(self%dat,1)
  ELSE
    hbw = 0   !> if not initialized, there are no bands
  END IF

END FUNCTION bsm_half_bw_


! ------------------------------------------------------------------------
!>  \brief Getter for matrix entries (non-exception checking)
!!
!!  \param  self  Reference to the banded symmetric matrix object
!!  \param  i     Row index (in unpacked version)
!!  \param  j     Column index (in unpacked version)
!!  \return v     Value at location (i,j) of the matrix
!!
!!  This routine determines the value at a particular location in the
!!  matrix. If the location is outside the band, it returns 0.
! ------------------------------------------------------------------------
FUNCTION bsm_get_ (self, i,j) RESULT(v)
  TYPE(bandSymMatrixT), INTENT(IN) :: self
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION :: v
  INTEGER :: r,c              !> indices in packed format

  !> get packed indices
  CALL packed_index(self, i,j, r,c)

  !> if desired location is within the bands, return the value
  IF (r.GT.0) THEN
    v = self%dat(r,c)
  ELSE
    v = 0.d0    !> if off-band, return 0
  END IF

END FUNCTION bsm_get_


! ------------------------------------------------------------------------
!>  \brief Getter for matrix entries (exception checking)
!!
!!  \param  self  Reference to the banded symmetric matrix object
!!  \param  i     Row index (in unpacked version)
!!  \param  j     Column index (in unpacked version)
!!  \param  exc   Error code
!!  \return v     Value at location (i,j) of the matrix
!!
!!  This routine determines the value at a particular location in the
!!  matrix. If the location is outside the band, it returns 0. If the
!!  location is not inside the matrix, it returns a POSIT exception.
! ------------------------------------------------------------------------
FUNCTION bsm_get_exc_ (self, i,j, exc) RESULT(v)
  TYPE(bandSymMatrixT), INTENT(IN) :: self
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: v
  INTEGER :: n    !> number of rows

  !> get number of rows (for efficiency)
  n = bsm_numRows(self)

  !> make sure desired indices are inside the matrix
  IF (i.GT.n .OR. i.LE.0 &
      .OR. j.GT.n .OR. j.LE.0) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    v = 0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception getter
  v = bsm_get(self, i,j)

END FUNCTION bsm_get_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for matrix entries (non-exception checking)
!!
!!  \param self Reference to the banded symmetric matrix object
!!  \param i    Row index (in unpacked version)
!!  \param j    Column index (in unpacked version)
!!  \param v    Value to be stored at location (i,j) of the matrix
!!
!!  This routine sets the value at a particular location in the
!!  matrix.
! ------------------------------------------------------------------------
SUBROUTINE bsm_set_ (self, i,j,v)
  TYPE(bandSymMatrixT), INTENT(INOUT) :: self
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: v
  INTEGER :: r,c        !> indices in packed format

  !> get packed indices
  CALL packed_index(self, i,j, r,c)

  !> set the data value
  self%dat(r,c) = v

END SUBROUTINE bsm_set_


! ------------------------------------------------------------------------
!>  \brief Setter for matrix entries (exception checking)
!!
!!  \param self Reference to the banded symmetric matrix object
!!  \param i    Row index (in unpacked version)
!!  \param j    Column index (in unpacked version)
!!  \param v    Value to be stored at location (i,j) of the matrix
!!  \param exc  Error code
!!
!!  This routine sets the value at a particular location in the
!!  matrix. If the location is outside the band or if the
!!  location is not inside the matrix, it returns a POSIT exception.
! ------------------------------------------------------------------------
SUBROUTINE bsm_set_exc_ (self, i,j,v, exc)
  TYPE(bandSymMatrixT), INTENT(INOUT) :: self
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: v
  INTEGER, INTENT(OUT) :: exc
  INTEGER :: n    !> number of rows

  !> get number of rows (for efficiency)
  n = bsm_numRows(self)

  !> ensure that desired indices are within the matrix
  !> and furthermore within the bands
  IF (i.GT.n .OR. i.LE.0 &
      .OR. j.GT.n .OR. j.LE.0 &
      .OR. ABS(j-i).GE.bsm_halfBW(self)) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL bsm_set(self, i,j,v)

END SUBROUTINE bsm_set_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for Cholesky decomposition (non-exception checking)
!!
!!  \param self   Reference to the banded symmetric matrix object
!!  \param decomp Array containing the decomposition in packed format
!!
!!  This routine sets the Cholesky decomposition of the matrix object.
!!  The decomposition is typically determined using the DPBTRF subroutine
!!  from the LAPACK library.
! ------------------------------------------------------------------------
SUBROUTINE bsm_set_decomp_ (self,decomp)
  TYPE(bandSymMatrixT), INTENT(INOUT) :: self
  DOUBLE PRECISION, INTENT(IN) :: decomp(:,:)

  !> set the decomposition data and flip the boolean
  self%decomp = decomp
  self%is_decomposed = .TRUE.

END SUBROUTINE bsm_set_decomp_


! ------------------------------------------------------------------------
!>  \brief Setter for Cholesky decomposition (exception checking)
!!
!!  \param self   Reference to the banded symmetric matrix object
!!  \param decomp Array containing the decomposition in packed format
!!  \param exc    Error code
!!
!!  This routine sets the Cholesky decomposition of the matrix object.
!!  The decomposition is typically determined using the DPBTRF subroutine
!!  from the LAPACK library. If the dimensions of the provided
!!  decomposition array do not match those of the allocated array, it
!!  returns a DIMEN exception.
! ------------------------------------------------------------------------
SUBROUTINE bsm_set_decomp_exc_ (self,decomp, exc)
  TYPE(bandSymMatrixT), INTENT(INOUT) :: self
  DOUBLE PRECISION, INTENT(IN) :: decomp(:,:)
  INTEGER, INTENT(OUT) :: exc

  !> make sure that the dimensions of the proposed decomposition match
  !> the actual dimensions of the packed storage
  IF (bsm_halfBW(self).NE.SIZE(decomp,1) &
        .OR. bsm_numRows(self).NE.SIZE(decomp,2)) THEN
    exc=DIMEN
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> call the non-exception version
  CALL bsm_setDecomp(self,decomp)

END SUBROUTINE bsm_set_decomp_exc_


! ------------------------------------------------------------------------
!>  \brief Getter for boolean indicating presence of decomposition
!!
!!  \param  self          Reference to the banded symmetric matrix object
!!  \return is_decomposed Boolean indicating whether the decomposition is present
! ------------------------------------------------------------------------
FUNCTION bsm_is_decomposed_ (self) RESULT(is_decomposed)
  TYPE(bandSymMatrixT), INTENT(IN) :: self
  LOGICAL :: is_decomposed

  is_decomposed = self%is_decomposed

END FUNCTION bsm_is_decomposed_


! ------------------------------------------------------------------------
!>  \brief Add two bandSymMatrixT objects (non-exception checking)
!!
!!  \param  self  Reference to the first (left-hand) banded symmetric matrix object
!!  \param  other Reference to the second (right-hand) banded symmetric matrix object
!!  \return new   Reference to the resulting banded symmetric matrix object
!!
!!  This routine adds two banded symmetric matrix objects. The half
!!  bandwidth of the result is the maximum of the half bandwidths of the
!!  input matrix objects.
! ------------------------------------------------------------------------
FUNCTION bsm_add_ (self,other) RESULT(new)
  TYPE(bandSymMatrixT), INTENT(IN) :: self,other
  TYPE(bandSymMatrixT) :: new
  INTEGER :: hbw1,hbw2    !> half bandwidths of inputs
  INTEGER :: hbw,n        !> half bandwidth of output and number of rows

  !> get number of rows and half bandwidths
  n = bsm_numRows(self)
  hbw1 = bsm_halfBW(self)
  hbw2 = bsm_halfBW(other)

  !> new half bandwidth is the maximum of the previous
  hbw = MAX(hbw1,hbw2)

  !> intialize new bandSymMatrixT
  CALL bsm_init(new, hbw,n)

  !> add the two matrices (using array ops for efficiency)
  new%dat(hbw-hbw1+1:hbw,:) = self%dat(:,:)
  new%dat(hbw-hbw2+1:hbw,:) = new%dat(hbw-hbw2+1:hbw,:) + other%dat(:,:)

END FUNCTION bsm_add_


! ------------------------------------------------------------------------
!>  \brief Add two bandSymMatrixT objects (exception checking)
!!
!!  \param  self  Reference to the first (left-hand) banded symmetric matrix object
!!  \param  other Reference to the second (right-hand) banded symmetric matrix object
!!  \param  exc   Error code
!!  \return new   Reference to the resulting banded symmetric matrix object
!!
!!  This routine adds two banded symmetric matrix objects. The half
!!  bandwidth of the result is the maximum of the half bandwidths of the
!!  input matrix objects. If the dimensions of the two input matrices do
!!  not match, it returns a DIMEN exception.
! ------------------------------------------------------------------------
FUNCTION bsm_add_exc_ (self,other, exc) RESULT(new)
  TYPE(bandSymMatrixT), INTENT(IN) :: self,other
  INTEGER, INTENT(OUT) :: exc
  TYPE(bandSymMatrixT) :: new

  !> make sure that the two matrices have the same number of rows
  IF (bsm_numRows(self).NE.bsm_numRows(other)) THEN
    exc=DIMEN
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  new = bsm_add(self,other)

END FUNCTION bsm_add_exc_


! ------------------------------------------------------------------------
!>  \brief Add a matrixT to a bandSymMatrixT in place with mapping (non-exception checking)
!!
!!  \param self   Reference to the banded symmetric matrix object
!!  \param other  Reference to the dense matrix object
!!  \param ind    Integer mapping indices
!!
!!  This routine adds a dense matrix to a banded symmetric matrix in place
!!  according to an indexed mapping. That is, the element at (i,j) in
!!  other is added to the element at (ind(i),ind(j)) of self.
! ------------------------------------------------------------------------
SUBROUTINE bsm_mapped_add_ (self, other,ind)
  TYPE(bandSymMatrixT), INTENT(INOUT) :: self
  TYPE(matrixT), INTENT(IN) :: other
  INTEGER, INTENT(IN) :: ind(:)
  INTEGER :: i,j    !> loop variables

  !> Add matrix entries corresponding to mapping indices
  DO i = 1,SIZE(ind)

    IF (ind(i).EQ.0)  CYCLE

    DO j = i,SIZE(ind)

      IF (ind(j).EQ.0) CYCLE

      CALL bsm_set(self, ind(i),ind(j), &
                    bsm_get(self,ind(i),ind(j)) &
                    + dm_get(other,i,j) )

    END DO
  END DO

END SUBROUTINE bsm_mapped_add_


! ------------------------------------------------------------------------
!>  \brief Add a matrixT to a bandSymMatrixT in place with mapping (exception checking)
!!
!!  \param self   Reference to the banded symmetric matrix object
!!  \param other  Reference to the dense matrix object
!!  \param ind    Integer mapping indices
!!  \param exc    Error code
!!
!!  This routine adds a dense matrix to a banded symmetric matrix in place
!!  according to an indexed mapping. That is, the element at (i,j) in
!!  other is added to the element at (ind(i),ind(j)) of self. If the
!!  dense matrix is not square or if the its dimensions exceed the half
!!  bandwidth of the banded symmetric matrix, it returns a DIMEN
!!  exception. If the number of mapping indices does not match the
!!  dimension of the dense matrix, it returns a DIMEN exception. If any of
!!  the index pairs indicate a location outside the matrix bands, it
!!  returns a POSIT exception. Note that a zero index is valid and
!!  ind(i)=0 or ind(j)=0 indicate that the value at (i,j) in the dense
!!  matrix object will not be added to the banded symmetric matrix object.
! ------------------------------------------------------------------------
SUBROUTINE bsm_mapped_add_exc_ (self, other,ind, exc)
  TYPE(bandSymMatrixT), INTENT(INOUT) :: self
  TYPE(matrixT), INTENT(IN) :: other
  INTEGER, INTENT(IN) :: ind(:)
  INTEGER, INTENT(OUT) :: exc

  !> ensure that matrixT is square (it is assumed to be symmetric)
  !> and that the number of rows in the matrixT matches the number
  !> of mapping indices
  !> and that the matrixT is not wider than the half bandwidth
  !> of the bandSymMatrixT
  IF (dm_numRows(other).NE.dm_numCols(other) &
      .OR. dm_numRows(other).NE.SIZE(ind) &
      .OR. dm_numRows(other).GT.bsm_halfBW(self)) THEN
    exc=DIMEN
    CALL log_printLogMsg(exc,sdr)
    RETURN

  !> ensure that there are no mapping indices outside the bandSymMatrixT
  !> and that all indices are within the bands
  ELSEIF ( MAXVAL(ind).GT.bsm_numRows(self) .OR. MINVAL(ind).LT.0 &
            .OR. MAXVAL(ind)-MINVAL(ind,MASK=ind.GT.0) .GE. bsm_halfBW(self) ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL bsm_mappedAdd(self,other,ind)

END SUBROUTINE bsm_mapped_add_exc_


! ------------------------------------------------------------------------
!>  \brief Scalar multiplication (self*k version)
!!
!!  \param  self  Reference to the banded symmetric matrix object
!!  \param  k     Scalar multiplication factor
!!  \return new   Reference to the resulting banded symmetric matrix object
!!
!!  This routine multiplies a banded symmetric matrix by a scalar factor.
! ------------------------------------------------------------------------
FUNCTION bsm_scal_mul_mat_scal_ (self,k) RESULT(new)
  TYPE(bandSymMatrixT), INTENT(IN) :: self
  DOUBLE PRECISION, INTENT(IN) :: k
  TYPE(bandSymMatrixT) :: new

  !> initialize output
  CALL bsm_init(new, bsm_halfBW(self),bsm_numRows(self))

  !> perform multiplication
  new%dat = k * self%dat

END FUNCTION bsm_scal_mul_mat_scal_


! ------------------------------------------------------------------------
!>  \brief Scalar multiplication (k*self version)
!!
!!  \param  self  Reference to the banded symmetric matrix object
!!  \param  k     Scalar multiplication factor
!!  \return new   Reference to the resulting banded symmetric matrix object
!!
!!  This routine multiplies a banded symmetric matrix by a scalar factor.
!!  The overload is necessary for the OPERATOR (*) overload. This version
!!  simply calls the other version for better maintainability.
! ------------------------------------------------------------------------
FUNCTION bsm_scal_mul_scal_mat_ (k,self) RESULT(new)
  TYPE(bandSymMatrixT), INTENT(IN) :: self
  DOUBLE PRECISION, INTENT(IN) :: k
  TYPE(bandSymMatrixT) :: new

  !> call self*k version
  new = bsm_scalMul(self,k)

END FUNCTION bsm_scal_mul_scal_mat_


! ------------------------------------------------------------------------
!>  \brief Post-multiplication of a banded symmetric matrix by a vector (non-exception checking)
!!
!!  \param  self  Reference to the banded symmetric matrix object
!!  \param  other Reference to the vector object
!!  \return new   Reference to the resulting vector object
!!
!!  This routine post-multiplies a banded symmetric matrix by a vector.
!!  That is, it performs self*other, not other*self.
! ------------------------------------------------------------------------
FUNCTION bsm_vec_mul_ (self,other) RESULT(new)
  TYPE(bandSymMatrixT), INTENT(IN) :: self
  TYPE(vectorT), INTENT(IN) :: other
  TYPE(vectorT) :: new
  INTEGER :: hbw,n    !> half bandwidth and number of rows
  INTEGER :: i,j      !> loop variables

  !> get size parameters
  n = vec_length(other)
  hbw = bsm_halfBW(self)

  !> initialize solution
  CALL vec_init(new, n)

  !> perform partial dot products of rows in self with the vector
  !> (ignoring off-band zeros)
  DO i = 1,n
    DO j = MAX(i-hbw+1,1), MIN(i+hbw-1,n)
      CALL vec_set(new, i, vec_get(new,i) + bsm_get(self,i,j)*vec_get(other,j))
    END DO
  END DO

END FUNCTION bsm_vec_mul_


! ------------------------------------------------------------------------
!>  \brief Post-multiplication of a banded symmetric matrix by a vector (exception checking)
!!
!!  \param  self  Reference to the banded symmetric matrix object
!!  \param  other Reference to the vector object
!!  \param  exc   Error code
!!  \return new   Reference to the resulting vector object
!!
!!  This routine post-multiplies a banded symmetric matrix by a vector.
!!  That is, it performs self*other, not other*self. If the dimensions
!!  of the two objects are not compatible, it returns a DIMEN exception.
! ------------------------------------------------------------------------
FUNCTION bsm_vec_mul_exc_ (self,other, exc) RESULT(new)
  TYPE(bandSymMatrixT), INTENT(IN) :: self
  TYPE(vectorT), INTENT(IN) :: other
  INTEGER, INTENT(OUT) :: exc
  TYPE(vectorT) :: new

  !> ensure that the number of rows in the bandSymMatrixT and
  !> the vectorT match
  IF (bsm_numRows(self).NE.vec_length(other)) THEN
    exc=DIMEN
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  new = bsm_vecMul(self,other)

END FUNCTION bsm_vec_mul_exc_



! ************************************************************************
! LOCAL FUNCTIONS
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Determine the packed indices from the non-packed indices
!!
!!  \param self Reference to the banded symmetric matrix object
!!  \param i    Row index (non-packed version)
!!  \param j    Column index (non-packed version)
!!  \param r    Row index (packed version)
!!  \param c    Column index (packed version)
!!
!!  This routine determines the indices of the desired location in the
!!  matrix in the packed data structure. It accounts for symmetry by
!!  ensuring that i <= j. After that is checked, the packed indices are
!!  computed as:  r = hbw - (j-i)
!!                c = j
! ------------------------------------------------------------------------
SUBROUTINE packed_index (self, i,j, r,c)
  TYPE(bandSymMatrixT), INTENT(IN) :: self
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: r,c

  !> flip indices - symmetry
  IF (i.GT.j) THEN
    r=j
    c=i
  ELSE
    r=i
    c=j
  END IF

  !> convert to packed index
  r = bsm_halfBW(self) - (c-r)

END SUBROUTINE packed_index

END MODULE band_sym_matrix_def
