! ------------------------------------------------------------------------
!>  \brief Module for testing Boundary Data module
! ------------------------------------------------------------------------
MODULE boundary_data_test
USE fruit                   !> Unit testing framework
USE system_constants        !> Global constants
USE log_message_control     !> Printing log/error messages
USE log_messages            !> Log/error codes
USE field_data              !> Field Data module
USE boundary_data           !> Boundary Data module

IMPLICIT NONE


CONTAINS


! ------------------------------------------------------------------------
!>  \test Test for OK exception message on allocation of elements
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nel        Number of elements
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!
!!  This test initializes the elements state variable in the boundary_data
!!  module and makes sure that the exception message is OK
!!  (i.e. allocation did not fail).
! ------------------------------------------------------------------------
SUBROUTINE test_bnd_elem_allocation_MSG
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_bnd_elem_allocation_MSG'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nel=100
  INTEGER, PARAMETER :: expMsg=OK
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log message file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize module and check the exception
  CALL bnd_init(nel, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate module
  CALL log_closeLogFile()
  CALL bnd_clean()

END SUBROUTINE test_bnd_elem_allocation_MSG


! ------------------------------------------------------------------------
!>  \test Test for SZE exception message on allocation of traction elements
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!
!!  This test attempts to initialize the elements state variable in the
!!  boundary data module with invalid size parameters and verifies that
!!  the correct exception is returned.
! ------------------------------------------------------------------------
SUBROUTINE test_bnd_elem_allocation_SZE
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_bnd_elem_allocation_SZE'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: expMsg=SZE
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> try to initialize with nel=0
  CALL bnd_init(0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> ensure module is reset
  CALL bnd_clean()

  !> try to initialize with nel=MAX_BOUNDELS+1
  CALL bnd_init(MAX_BOUNDELS+1, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate module
  CALL log_closeLogFile()
  CALL bnd_clean()

END SUBROUTINE test_bnd_elem_allocation_SZE


! ------------------------------------------------------------------------
!>  \test Test for number of elements when element data is not initialized
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param expected     Expected number of elements
!!  \param actual       Actual number of elements
!!
!!  This test makes sure that the number of elements is returned as 0
!!  when the element data is not initialized
! ------------------------------------------------------------------------
SUBROUTINE test_bnd_num_elem_not_allocated
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_bnd_num_elem_not_allocated'
  INTEGER, PARAMETER :: expected = 0
  INTEGER :: actual

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> check number of materials
  actual = bnd_numBoundElem()
  CALL assertEquals(expected, actual)

END SUBROUTINE test_bnd_num_elem_not_allocated


! ------------------------------------------------------------------------
!>  \test Test for number of elements when element data is initialized
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param expected     Expected number of elements
!!  \param actual       Actual number of elements
!!
!!  This test makes sure that the correct number of elements is returned.
! ------------------------------------------------------------------------
SUBROUTINE test_bnd_num_elem_allocated
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_bnd_num_elem_allocated'
  INTEGER, PARAMETER :: expected = 100
  INTEGER :: actual

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize the module
  CALL bnd_init(expected)

  !> check number of nodes
  actual = bnd_numBoundElem()
  CALL assertEquals(expected, actual)

  !> deallocate the module
  CALL bnd_clean()

END SUBROUTINE test_bnd_num_elem_allocated


! ------------------------------------------------------------------------
!>  \test Test for POSIT exception from bnd_getConnect
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nel        Number of elements
!!  \param iel        Element number
!!  \param jnod       Local node index
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!  \param n          Dummy variable for get function return
!!
!!  This test checks that a POSIT exception is returned when the requested
!!  location is not inside the element list or the requested local node
!!  index is not within [1..NNODELB].
! ------------------------------------------------------------------------
SUBROUTINE test_bnd_get_connect_POSIT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_bnd_get_connect_POSIT'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nel=100
  INTEGER, PARAMETER :: iel = 50
  INTEGER, PARAMETER :: jnod = 2
  INTEGER, PARAMETER :: expMsg = POSIT
  INTEGER :: actMsg
  INTEGER :: n

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize nodes
  CALL bnd_init(nel)

  !> try to get beyond last element
  n = bnd_getConnect(nel+1,jnod, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to get before first element
  n = bnd_getConnect(0,jnod, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to get beyond last local node index
  n = bnd_getConnect(iel,NNODELB+1, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to get before first local node index
  n = bnd_getConnect(iel,0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate the elements list
  CALL log_closeLogFile()
  CALL bnd_clean()

END SUBROUTINE test_bnd_get_connect_POSIT


! ------------------------------------------------------------------------
!>  \test Test for POSIT exception from bnd_setConnect
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nnod       Number of nodes
!!  \param nel        Number of elements
!!  \param iel        Element number
!!  \param jnod       Local node index
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!  \param n          Dummy variable for set routine input
!!
!!  This test checks that a POSIT exception is returned when the requested
!!  location is not inside the element list or the requested local node
!!  index is not within [1..NNODELB].
! ------------------------------------------------------------------------
SUBROUTINE test_bnd_set_connect_POSIT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_bnd_set_connect_POSIT'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nnod=100
  INTEGER, PARAMETER :: nel=100
  INTEGER, PARAMETER :: iel = 50
  INTEGER, PARAMETER :: jnod = 2
  INTEGER, PARAMETER :: expMsg = POSIT
  INTEGER :: actMsg
  INTEGER, PARAMETER :: n = 10

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize nodes and elements
  CALL bnd_init(nel)

  !> try to set beyond last element
  CALL bnd_setConnect(nel+1,jnod, n, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set before first element
  CALL bnd_setConnect(0,jnod, n, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set beyond last local node index
  CALL bnd_setConnect(iel,NNODELB+1, n, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set before first local node index
  CALL bnd_setConnect(iel,0, n, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate elements
  CALL log_closeLogFile()
  CALL bnd_clean()

END SUBROUTINE test_bnd_set_connect_POSIT


! ------------------------------------------------------------------------
!>  \test Test for EXCEED exception from bnd_setConnect
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nnod       Number of nodes
!!  \param nel        Number of elements
!!  \param iel        Element number
!!  \param jnod       Local node index
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!
!!  This test checks that an EXCEED exception is returned when the
!!  specified input is not within the range defined in the System
!!  Constants module.
! ------------------------------------------------------------------------
SUBROUTINE test_bnd_set_connect_EXCEED
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_bnd_set_connect_EXCEED'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nnod=100
  INTEGER, PARAMETER :: nel=100
  INTEGER, PARAMETER :: iel = 50
  INTEGER, PARAMETER :: jnod = 2
  INTEGER, PARAMETER :: expMsg = EXCEED
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize nodes and elements
  CALL fld_initNode(nnod)
  CALL bnd_init(nel)

  !> try to set below minimum
  CALL bnd_setConnect(iel,jnod, 0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set above maximum
  CALL bnd_setConnect(iel,jnod, nnod+1, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate the nodes and elements
  CALL log_closeLogFile()
  CALL fld_cleanNode()
  CALL bnd_clean()

END SUBROUTINE test_bnd_set_connect_EXCEED


! ------------------------------------------------------------------------
!>  \test Test for correct value getting and setting in bnd_getConnect
!!        and bnd_setConnect
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param nnod       Number of nodes
!!  \param nel        Number of elements
!!  \param iel        Element number
!!  \param jnod       Local node index
!!  \param expVal     Expected data value
!!  \param actVal     Actual data value
!!
!!  This test checks that the correct value is set using the
!!  bnd_setConnect access program and returned from the
!!  bnd_getConnect access program.
! ------------------------------------------------------------------------
SUBROUTINE test_bnd_get_set_connect_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_bnd_get_set_connect_VAL'
  INTEGER, PARAMETER :: nnod=100
  INTEGER, PARAMETER :: nel=100
  INTEGER, PARAMETER :: iel = 50
  INTEGER, PARAMETER :: jnod = 2
  INTEGER, PARAMETER :: expVal = 20
  INTEGER :: actVal

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize nodes and elements
  CALL fld_initNode(nnod)
  CALL bnd_init(nel)

  !> set connectivity value
  CALL bnd_setConnect(iel,jnod, expVal)
  actVal = bnd_getConnect(iel,jnod)
  CALL assertEquals(expVal,actVal)

  !> finalize the log file and deallocate the nodes and elements
  CALL log_closeLogFile()
  CALL fld_cleanNode()
  CALL bnd_clean()

END SUBROUTINE test_bnd_get_set_connect_VAL


! ------------------------------------------------------------------------
!>  \test Test for POSIT exception message from bnd_lenBoundElem
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nel        Number of elements
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!  \param length     Dummy variable for get function return
!!
!!  This test makes sure that the correct exception is returned when the
!!  specified element index is not in the range of the allocated element
!!  data.
! ------------------------------------------------------------------------
SUBROUTINE test_bnd_len_bound_elem_POSIT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_bnd_len_bound_elem_POSIT'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nel=100
  INTEGER, PARAMETER :: expMsg=POSIT
  INTEGER :: actMsg
  DOUBLE PRECISION :: length

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize elements
  CALL bnd_init(nel)

  !> try to get beyond last element
  length = bnd_lenBoundElem(nel+1, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to get before first element
  length = bnd_lenBoundElem(0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate the elements list
  CALL log_closeLogFile()
  CALL bnd_clean()

END SUBROUTINE test_bnd_len_bound_elem_POSIT


! ------------------------------------------------------------------------
!>  \test Test for element length when node coordinates not initialized
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param nnod       Number of nodes
!!  \param nel        Number of elements
!!  \param iel        Element number
!!  \param expected   Expected length of element (zero)
!!  \param actual     Actual length of element
!!
!!  This test makes sure that the length of the element is returned as 0
!!  when the coordinates of the connected nodes have not been set
! ------------------------------------------------------------------------
SUBROUTINE test_bnd_len_bound_elem_ZERO
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_bnd_len_bound_elem_ZERO'
  INTEGER, PARAMETER :: nnod=100
  INTEGER, PARAMETER :: nel=100
  DOUBLE PRECISION, PARAMETER :: expected = 0.d0
  DOUBLE PRECISION :: actual

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize nodes and elements
  CALL fld_initNode(nnod)
  CALL bnd_init(nel)

  !> set connectivity of first element to {1,2}
  CALL bnd_setConnect(1,1,1)
  CALL bnd_setConnect(1,2,2)

  !> check length of element
  actual = bnd_lenBoundElem(1)
  CALL assertEquals(expected,actual)

  !> deallocate the elements and nodes
  CALL fld_cleanNode()
  CALL bnd_clean()

END SUBROUTINE test_bnd_len_bound_elem_ZERO


! ------------------------------------------------------------------------
!>  \test Test for element length when node coordinates initialized
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param nnod       Number of nodes
!!  \param nel        Number of elements
!!  \param iel        Element number
!!  \param expected   Expected length of element
!!  \param actual     Actual length of element
!!
!!  This test makes sure that the length of the element is computed
!!  correctly for the line given by (1,1), (3,2)
! ------------------------------------------------------------------------
SUBROUTINE test_bnd_len_bound_elem_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_bnd_len_bound_elem_VAL'
  INTEGER, PARAMETER :: nnod=100
  INTEGER, PARAMETER :: nel=100
  DOUBLE PRECISION, PARAMETER :: toler = 1.d-14
  DOUBLE PRECISION, PARAMETER :: expected = 2.23606797749979d0
  DOUBLE PRECISION :: actual

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize nodes and elements
  CALL fld_initNode(nnod)
  CALL bnd_init(nel)

  !> set connectivity of first element to {1,2}
  CALL bnd_setConnect(1,1,1)
  CALL bnd_setConnect(1,2,2)

  !> set coordinates of nodes
  CALL fld_setCoord(1,1, 1.d0)
  CALL fld_setCoord(1,2, 1.d0)
  CALL fld_setCoord(2,1, 3.d0)
  CALL fld_setCoord(2,2, 2.d0)

  !> check volume of element
  actual = bnd_lenBoundElem(1)
  CALL assertEquals(expected,actual, toler)

  !> deallocate the elements and nodes
  CALL fld_cleanNode()
  CALL bnd_clean()

END SUBROUTINE test_bnd_len_bound_elem_VAL


! ------------------------------------------------------------------------
!>  \test Test for POSIT exception from bnd_getTrac
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nel        Number of elements
!!  \param iel        Element number
!!  \param jnod       Local node index
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!  \param load       Dummy variable for get function return
!!
!!  This test checks that a POSIT exception is returned when the requested
!!  location is not inside the element list or the requested local node
!!  index is not within [1..NNODELB].
! ------------------------------------------------------------------------
SUBROUTINE test_bnd_get_trac_POSIT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_bnd_get_trac_POSIT'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nel=100
  INTEGER, PARAMETER :: iel = 50
  INTEGER, PARAMETER :: jnod = 2
  INTEGER, PARAMETER :: expMsg = POSIT
  INTEGER :: actMsg
  TYPE(surfLoadT) :: load

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize nodes
  CALL bnd_init(nel)

  !> try to get beyond last element
  load = bnd_getTrac(nel+1,jnod, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to get before first element
  load = bnd_getTrac(0,jnod, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to get beyond last local node index
  load = bnd_getTrac(iel,NNODELB+1, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to get before first local node index
  load = bnd_getTrac(iel,0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate the elements list
  CALL log_closeLogFile()
  CALL bnd_clean()

END SUBROUTINE test_bnd_get_trac_POSIT


! ------------------------------------------------------------------------
!>  \test Test for POSIT exception from bnd_setTrac
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nnod       Number of nodes
!!  \param nel        Number of elements
!!  \param iel        Element number
!!  \param jnod       Local node index
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!  \param load       Dummy variable for set routine input
!!
!!  This test checks that a POSIT exception is returned when the requested
!!  location is not inside the element list or the requested local node
!!  index is not within [1..NNODELB].
! ------------------------------------------------------------------------
SUBROUTINE test_bnd_set_trac_POSIT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_bnd_set_trac_POSIT'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nnod=100
  INTEGER, PARAMETER :: nel=100
  INTEGER, PARAMETER :: iel = 50
  INTEGER, PARAMETER :: jnod = 2
  INTEGER, PARAMETER :: expMsg = POSIT
  INTEGER :: actMsg
  TYPE(surfLoadT) :: load

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize elements
  CALL bnd_init(nel)

  !> try to set beyond last element
  CALL bnd_setTrac(nel+1,jnod, load, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set before first element
  CALL bnd_setTrac(0,jnod, load, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set beyond last local node index
  CALL bnd_setTrac(iel,NNODELB+1, load, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set before first local node index
  CALL bnd_setTrac(iel,0, load, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate the nodes and elements
  CALL log_closeLogFile()
  CALL bnd_clean()

END SUBROUTINE test_bnd_set_trac_POSIT


! ------------------------------------------------------------------------
!>  \test Test for EXCEED exception from bnd_setTrac
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nel        Number of elements
!!  \param iel        Element number
!!  \param jnod       Local node index
!!  \param sig_nt     Test shear stress
!!  \param sig_nn     Test normal stress
!!  \param load       Surface load
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!
!!  This test checks that an EXCEED exception is returned when the
!!  specified input is not within the range defined in the System
!!  Constants module.
! ------------------------------------------------------------------------
SUBROUTINE test_bnd_set_trac_EXCEED
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_bnd_set_trac_EXCEED'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nel=100
  INTEGER, PARAMETER :: iel = 50
  INTEGER, PARAMETER :: jnod = 2
  DOUBLE PRECISION :: sig_nt, sig_nn
  TYPE(surfLoadT) :: load
  INTEGER, PARAMETER :: expMsg = EXCEED
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize elements
  CALL bnd_init(nel)

  !> set load such that sig_nt < SIG_MIN
  sig_nt = SIG_MIN-0.1d0
  sig_nn = 0.d0
  load%sig_nt = sig_nt
  load%sig_nn = sig_nn

  !> try to set traction
  CALL bnd_setTrac(iel,jnod, load, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> set load such that sig_nt > SIG_MAX
  sig_nt = SIG_MAX+0.1d0
  load%sig_nt = sig_nt

  !> try to set traction
  CALL bnd_setTrac(iel,jnod, load, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> set load such that sig_nn < SIG_MIN
  sig_nt = 0.d0
  sig_nn = SIG_MIN-0.1d0
  load%sig_nt = sig_nt
  load%sig_nn = sig_nn

  !> try to set traction
  CALL bnd_setTrac(iel,jnod, load, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> set load such that sig_nn > SIG_MAX
  sig_nn = SIG_MAX+0.1d0
  load%sig_nn = sig_nn

  !> try to set traction
  CALL bnd_setTrac(iel,jnod, load, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate elements
  CALL log_closeLogFile()
  CALL bnd_clean()

END SUBROUTINE test_bnd_set_trac_EXCEED


! ------------------------------------------------------------------------
!>  \test Test for correct value getting and setting in bnd_getTrac
!!        and bnd_setTrac
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param nel        Number of elements
!!  \param iel        Element number
!!  \param jnod       Local node index
!!  \param load       Surface load
!!  \param expSigNT   Expected shear stress
!!  \param expSigNN   Expected normal stress
!!  \param actSigNT   Actual shear stress
!!  \param actSigNN   Actual normal stress
!!
!!  This test checks that the correct value is set using the
!!  bnd_setTrac access program and returned from the
!!  bnd_getTrac access program.
! ------------------------------------------------------------------------
SUBROUTINE test_bnd_get_set_trac_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_bnd_get_set_trac_VAL'
  INTEGER, PARAMETER :: nnod=100
  INTEGER, PARAMETER :: nel=100
  INTEGER, PARAMETER :: iel = 50
  INTEGER, PARAMETER :: jnod = 2
  DOUBLE PRECISION, PARAMETER :: expSigNT=1.d3, expSigNN=-2.d3
  TYPE(surfLoadT) :: load
  DOUBLE PRECISION :: actSigNT, actSigNN

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize elements
  CALL bnd_init(nel)

  !> set traction value
  load%sig_nt = expSigNT
  load%sig_nn = expSigNN
  CALL bnd_setTrac(iel,jnod, load)

  !> reset load
  load%sig_nt = 0.d0
  load%sig_nn = 0.d0

  !> get traction value
  load = bnd_getTrac(iel,jnod)
  actSigNT = load%sig_nt
  actSigNN = load%sig_nn
  CALL assertEquals(expSigNT,actSigNT)
  CALL assertEquals(expSigNN,actSigNN)

  !> finalize the log file and deallocate the nodes and elements
  CALL log_closeLogFile()
  CALL bnd_clean()

END SUBROUTINE test_bnd_get_set_trac_VAL

END MODULE boundary_data_test
