! ------------------------------------------------------------------------
!>  \brief Module for Field Data
! ------------------------------------------------------------------------
MODULE field_data
USE system_constants      !> Global system constants
USE log_message_control   !> Printing log/error messages
USE log_messages          !> Log/error message and sender codes
USE material_data         !> Material Property Data

IMPLICIT NONE
PRIVATE



! ************************************************************************
! EXPORTS
! ************************************************************************


!> Exported interfaces
PUBLIC :: fld_initTime, fld_cleanTime, &
          fld_timeStep, fld_numTimeSteps, &
          fld_initNode, fld_cleanNode, &
          fld_numNode, &
          fld_getCoord, fld_setCoord, &
          fld_getFix, fld_setFix, &
          fld_initDof, fld_numDof, fld_getDof, &
          fld_getDisp, fld_setDisp, &
          fld_getVel, fld_setVel, &
          fld_getAcc, fld_setAcc, &
          fld_getBodyAcc, fld_setBodyAcc, &
          fld_getStressNode, fld_setStressNode, &
          fld_getStrainNode, fld_setStrainNode, &
          fld_initElem, fld_cleanElem, &
          fld_numElem, &
          fld_getConnect, fld_setConnect, &
          fld_volElem, &
          fld_getMaterial, fld_setMaterial, &
          fld_getStressElem, fld_setStressElem, &
          fld_getStrainElem, fld_setStrainElem



! ************************************************************************
! LOCAL CONSTANTS
! ************************************************************************

!> sender code for this module
INTEGER, PARAMETER :: sdr = FLDDAT

!> subscripts for tensor variables
INTEGER, PARAMETER, PUBLIC :: s11=1, s22=2, s12=3, s33=4



! ************************************************************************
! DATA TYPES
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Coordinate
!!
!!  \param x    Set of coordinates
! ------------------------------------------------------------------------
TYPE coordT
  DOUBLE PRECISION, DIMENSION(NDIM) :: x
END TYPE coordT


! ------------------------------------------------------------------------
!>  \brief Displacement
!!
!!  \param u    Set of displacements
! ------------------------------------------------------------------------
TYPE dispT
  DOUBLE PRECISION, DIMENSION(NDIM) :: u
END TYPE dispT


! ------------------------------------------------------------------------
!>  \brief Velocity
!!
!!  \param v    Set of velocities
! ------------------------------------------------------------------------
TYPE velT
  DOUBLE PRECISION, DIMENSION(NDIM) :: v
END TYPE velT


! ------------------------------------------------------------------------
!>  \brief Acceleration
!!
!!  \param a    Set of accelerations
! ------------------------------------------------------------------------
TYPE accT
  DOUBLE PRECISION, DIMENSION(NDIM) :: a
END TYPE accT


! ------------------------------------------------------------------------
!>  \brief Fixity
!!
!!  \param fix   Set of fixities
! ------------------------------------------------------------------------
TYPE fixT
  LOGICAL, DIMENSION(NDIM) :: fix
END TYPE fixT


! ------------------------------------------------------------------------
!>  \brief Degrees of freedom
!!
!!  \param dof   Set of degree of freedom indices
! ------------------------------------------------------------------------
TYPE dofT
  INTEGER, DIMENSION(NDIM) :: dof
END TYPE dofT


! ------------------------------------------------------------------------
!>  \brief Stress
!!
!!  \param sig   Set of stresses
! ------------------------------------------------------------------------
TYPE stressT
  DOUBLE PRECISION, DIMENSION(NTNS) :: sig
END TYPE stressT


! ------------------------------------------------------------------------
!>  \brief Strain
!!
!!  \param eps   Set of strains
! ------------------------------------------------------------------------
TYPE strainT
  DOUBLE PRECISION, DIMENSION(NTNS) :: eps
END TYPE strainT


! ------------------------------------------------------------------------
!>  \brief Connectivity
!!
!!  \param nd   Sequence of node indices
! ------------------------------------------------------------------------
TYPE connectT
  INTEGER, DIMENSION(NNODEL) :: nd
END TYPE connectT


! ------------------------------------------------------------------------
!>  \brief Node
!!
!!  \param num      Node number
!!  \param pt       Node coordinates
!!  \param fix      Fixities
!!  \param dof      Degrees of freedom
!!  \param disp     Displacement
!!  \param vel      Velocity
!!  \param acc      Acceleration
!!  \param body     Body acceleration
!!  \param stress   Nodal stress
!!  \param strain   Nodal strain
! ------------------------------------------------------------------------
TYPE nodeT
  INTEGER       :: num
  TYPE(coordT)  :: pt
  TYPE(fixT)    :: fix
  TYPE(dofT)    :: dof
  TYPE(dispT)   :: disp
  TYPE(velT)    :: vel
  TYPE(accT)    :: acc
  TYPE(accT)    :: body
  TYPE(stressT) :: stress
  TYPE(strainT) :: strain
END TYPE nodeT


! ------------------------------------------------------------------------
!>  \brief Element
!!
!!  \param num      Element number
!!  \param ico      Connectivity of local nodes
!!  \param mtl      Material number
!!  \param stress   Element stress
!!  \param strain   Element strain
! ------------------------------------------------------------------------
TYPE elementT
  INTEGER         :: num
  TYPE(connectT)  :: ico
  INTEGER         :: mtl
  TYPE(stressT)   :: stress
  TYPE(strainT)   :: strain
END TYPE elementT



! ************************************************************************
! STATE VARIABLES
! ************************************************************************

DOUBLE PRECISION, SAVE        :: dTime        = 0.d0    !> time step
INTEGER,          SAVE        :: nTime        = 0       !> number of time steps
TYPE(nodeT),      ALLOCATABLE :: nodes(:)               !> set of nodes
TYPE(elementT),   ALLOCATABLE :: elements(:)            !> set of body elements
INTEGER,          SAVE        :: ndof         = 0       !> total number of degrees of freedom



! ************************************************************************
! INTERFACES
! ************************************************************************


!> \brief Interface to initializer for time-stepping information
INTERFACE fld_initTime
  MODULE PROCEDURE fld_init_time_
  MODULE PROCEDURE fld_init_time_exc_
END INTERFACE fld_initTime

!> \brief Interface to destructor for time-stepping information
INTERFACE fld_cleanTime
  MODULE PROCEDURE fld_clean_time_
END INTERFACE fld_cleanTime

!> \brief Interface to getter for time step
INTERFACE fld_timeStep
  MODULE PROCEDURE fld_time_step_
END INTERFACE fld_timeStep

!> \brief Interface to getter for number of time steps
INTERFACE fld_numTimeSteps
  MODULE PROCEDURE fld_num_time_steps_
END INTERFACE fld_numTimeSteps

!> \brief Interface to initializer for node information
INTERFACE fld_initNode
  MODULE PROCEDURE fld_init_node_
  MODULE PROCEDURE fld_init_node_exc_
END INTERFACE fld_initNode

!> \brief Interface to destructor for node information
INTERFACE fld_cleanNode
  MODULE PROCEDURE fld_clean_node_
END INTERFACE fld_cleanNode

!> \brief Interface to getter for number of nodes
INTERFACE fld_numNode
  MODULE PROCEDURE fld_num_node_
END INTERFACE fld_numNode

!> \brief Interface to getter for node coordinates
INTERFACE fld_getCoord
  MODULE PROCEDURE fld_get_coord_
  MODULE PROCEDURE fld_get_coord_exc_
END INTERFACE fld_getCoord

!> \brief Interface to setter for node coordinates
INTERFACE fld_setCoord
  MODULE PROCEDURE fld_set_coord_
  MODULE PROCEDURE fld_set_coord_exc_
END INTERFACE fld_setCoord

!> \brief Interface to getter for node fixity
INTERFACE fld_getFix
  MODULE PROCEDURE fld_get_fix_
  MODULE PROCEDURE fld_get_fix_exc_
END INTERFACE fld_getFix

!> \brief Interface to setter for node fixity
INTERFACE fld_setFix
  MODULE PROCEDURE fld_set_fix_
  MODULE PROCEDURE fld_set_fix_exc_
END INTERFACE fld_setFix

!> \brief Interface to initializer for degrees of freedom
INTERFACE fld_initDof
  MODULE PROCEDURE fld_init_dof_exc_
END INTERFACE fld_initDof

!> \brief Interface to getter for number of degrees of freedom
INTERFACE fld_numDof
  MODULE PROCEDURE fld_num_dof_
END INTERFACE fld_numDof

!> \brief Interface to getter for degree of freedom
INTERFACE fld_getDof
  MODULE PROCEDURE fld_get_dof_
  MODULE PROCEDURE fld_get_dof_exc_
END INTERFACE fld_getDof

!> \brief Interface to getter for node displacement
INTERFACE fld_getDisp
  MODULE PROCEDURE fld_get_disp_
  MODULE PROCEDURE fld_get_disp_exc_
END INTERFACE fld_getDisp

!> \brief Interface to setter for node displacement
INTERFACE fld_setDisp
  MODULE PROCEDURE fld_set_disp_
  MODULE PROCEDURE fld_set_disp_exc_
END INTERFACE fld_setDisp

!> \brief Interface to getter for node velocity
INTERFACE fld_getVel
  MODULE PROCEDURE fld_get_vel_
  MODULE PROCEDURE fld_get_vel_exc_
END INTERFACE fld_getVel

!> \brief Interface to setter for node velocity
INTERFACE fld_setVel
  MODULE PROCEDURE fld_set_vel_
  MODULE PROCEDURE fld_set_vel_exc_
END INTERFACE fld_setVel

!> \brief Interface to getter for node acceleration
INTERFACE fld_getAcc
  MODULE PROCEDURE fld_get_acc_
  MODULE PROCEDURE fld_get_acc_exc_
END INTERFACE fld_getAcc

!> \brief Interface to setter for node acceleration
INTERFACE fld_setAcc
  MODULE PROCEDURE fld_set_acc_
  MODULE PROCEDURE fld_set_acc_exc_
END INTERFACE fld_setAcc

!> \brief Interface to getter for body acceleration at node
INTERFACE fld_getBodyAcc
  MODULE PROCEDURE fld_get_body_acc_
  MODULE PROCEDURE fld_get_body_acc_exc_
END INTERFACE fld_getBodyAcc

!> \brief Interface to setter for body acceleration at node
INTERFACE fld_setBodyAcc
  MODULE PROCEDURE fld_set_body_acc_
  MODULE PROCEDURE fld_set_body_acc_exc_
END INTERFACE fld_setBodyAcc

!> \brief Interface to getter for node stress
INTERFACE fld_getStressNode
  MODULE PROCEDURE fld_get_stress_node_
  MODULE PROCEDURE fld_get_stress_node_exc_
END INTERFACE fld_getStressNode

!> \brief Interface to setter for node stress
INTERFACE fld_setStressNode
  MODULE PROCEDURE fld_set_stress_node_
  MODULE PROCEDURE fld_set_stress_node_exc_
END INTERFACE fld_setStressNode

!> \brief Interface to getter for node strain
INTERFACE fld_getStrainNode
  MODULE PROCEDURE fld_get_strain_node_
  MODULE PROCEDURE fld_get_strain_node_exc_
END INTERFACE fld_getStrainNode

!> \brief Interface to setter for node strain
INTERFACE fld_setStrainNode
  MODULE PROCEDURE fld_set_strain_node_
  MODULE PROCEDURE fld_set_strain_node_exc_
END INTERFACE fld_setStrainNode

!> \brief Interface to initializer for elements
INTERFACE fld_initElem
  MODULE PROCEDURE fld_init_elem_
  MODULE PROCEDURE fld_init_elem_exc_
END INTERFACE fld_initElem

!> \brief Interface to destructor for element information
INTERFACE fld_cleanElem
  MODULE PROCEDURE fld_clean_elem_
END INTERFACE fld_cleanElem

!> \brief Interface to getter for number of elements
INTERFACE fld_numElem
  MODULE PROCEDURE fld_num_elem_
END INTERFACE fld_numElem

!> \brief Interface to getter for element connectivity
INTERFACE fld_getConnect
  MODULE PROCEDURE fld_get_connect_
  MODULE PROCEDURE fld_get_connect_exc_
END INTERFACE fld_getConnect

!> \brief Interface to setter for element connectivity
INTERFACE fld_setConnect
  MODULE PROCEDURE fld_set_connect_
  MODULE PROCEDURE fld_set_connect_exc_
END INTERFACE fld_setConnect

!> \brief Interface to getter for volume of element
INTERFACE fld_volElem
  MODULE PROCEDURE fld_vol_elem_
  MODULE PROCEDURE fld_vol_elem_exc_
END INTERFACE fld_volElem

!> \brief Interface to getter for element material
INTERFACE fld_getMaterial
  MODULE PROCEDURE fld_get_material_
  MODULE PROCEDURE fld_get_material_exc_
END INTERFACE fld_getMaterial

!> \brief Interface to setter for element material
INTERFACE fld_setMaterial
  MODULE PROCEDURE fld_set_material_
  MODULE PROCEDURE fld_set_material_exc_
END INTERFACE fld_setMaterial

!> \brief Interface to getter for element stress
INTERFACE fld_getStressElem
  MODULE PROCEDURE fld_get_stress_elem_
  MODULE PROCEDURE fld_get_stress_elem_exc_
END INTERFACE fld_getStressElem

!> \brief Interface to setter for element stress
INTERFACE fld_setStressElem
  MODULE PROCEDURE fld_set_stress_elem_
  MODULE PROCEDURE fld_set_stress_elem_exc_
END INTERFACE fld_setStressElem

!> \brief Interface to getter for element strain
INTERFACE fld_getStrainElem
  MODULE PROCEDURE fld_get_strain_elem_
  MODULE PROCEDURE fld_get_strain_elem_exc_
END INTERFACE fld_getStrainElem

!> \brief Interface to setter for element strain
INTERFACE fld_setStrainElem
  MODULE PROCEDURE fld_set_strain_elem_
  MODULE PROCEDURE fld_set_strain_elem_exc_
END INTERFACE fld_setStrainElem



CONTAINS



! ************************************************************************
! ACCESS PROGRAMS
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Initializer for time-stepping data (non-exception checking)
!!
!!  \param  dt  Time step
!!  \param  n   Number of time steps
!!
!!  This routine sets the module state parameters for time step and total
!!  number of time steps.
! ------------------------------------------------------------------------
SUBROUTINE fld_init_time_ (dt,n)
  DOUBLE PRECISION, INTENT(IN) :: dt
  INTEGER, INTENT(IN) :: n

  dTime = dt
  nTime = n

END SUBROUTINE fld_init_time_


! ------------------------------------------------------------------------
!>  \brief Initializer for time-stepping data (exception checking)
!!
!!  \param  dt  Time step
!!  \param  n   Number of time steps
!!  \param  exc Error code
!!
!!  \exception  EXCEED  Size of time step or number of time steps exceed definined min/max values
!!
!!  This routine sets the module state parameters for time step and total
!!  number of time steps.
! ------------------------------------------------------------------------
SUBROUTINE fld_init_time_exc_ (dt,n, exc)
  DOUBLE PRECISION, INTENT(IN) :: dt
  INTEGER, INTENT(IN) :: n
  INTEGER, INTENT(OUT) :: exc

  !> check that time step and number of time steps are valid
  IF (  dt.LT.DTIME_MIN .OR. dt.GT.DTIME_MAX  &
        .OR. n.LT.1 .OR. n.GT.MAX_TIMESTEPS ) THEN
    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    CALL fld_cleanTime()
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL fld_initTime(dt,n)

END SUBROUTINE fld_init_time_exc_


! ------------------------------------------------------------------------
!>  \brief Destructor for time-stepping data
!!
!!  This routine resets the module state parameters for time step and
!!  total number of time steps to 0.
! ------------------------------------------------------------------------
SUBROUTINE fld_clean_time_ ()

  dTime = 0.d0
  nTime = 0

END SUBROUTINE fld_clean_time_


! ------------------------------------------------------------------------
!>  \brief Getter for size of time step
!!
!!  \return dt  Time step
! ------------------------------------------------------------------------
FUNCTION fld_time_step_ () RESULT(dt)
  DOUBLE PRECISION :: dt

  dt = dTime

END FUNCTION fld_time_step_


! ------------------------------------------------------------------------
!>  \brief Getter for number of time steps
!!
!!  \return n   Number of time steps
! ------------------------------------------------------------------------
FUNCTION fld_num_time_steps_ () RESULT(n)
  DOUBLE PRECISION :: n

  n = nTime

END FUNCTION fld_num_time_steps_


! ------------------------------------------------------------------------
!>  \brief Constructor for node data (non-exception checking)
!!
!!  \param  nnod  Number of nodes
!!
!!  This routine allocates memory and initializes the state variable that
!!  contains the set of nodes for the problem.
! ------------------------------------------------------------------------
SUBROUTINE fld_init_node_ (nnod)
  INTEGER, INTENT(IN) :: nnod
  INTEGER :: i, j     !> loop variables

  !> only reallocate if new dimensions do not match existing dimensions
  IF (fld_numNode().NE.nnod) THEN

    !> ensure state variable is clear
    CALL fld_cleanNode()

    !> allocate memory for node data state variable
    ALLOCATE(nodes(nnod))

  END IF

  !> initialize state variable

  ndof = 0

  DO i = 1,nnod

    !> node number
    nodes(i)%num   = i

    !> vector variables
    DO j = 1,NDIM
      nodes(i)%pt   %x(j)   = 0.d0
      nodes(i)%fix  %fix(j) = .FALSE.
      nodes(i)%dof  %dof(j) = 0
      nodes(i)%disp %u(j)   = 0.d0
      nodes(i)%vel  %v(j)   = 0.d0
      nodes(i)%acc  %a(j)   = 0.d0
      nodes(i)%body %a(j)   = 0.d0
    END DO

    !> tensor variables
    DO j = 1,NTNS
      nodes(i)%stress%sig(j) = 0.d0
      nodes(i)%strain%eps(j) = 0.d0
    END DO

  END DO

END SUBROUTINE fld_init_node_


! ------------------------------------------------------------------------
!>  \brief Constructor for node data (exception checking)
!!
!!  \param  nnod  Number of nodes
!!  \param  exc   Error code
!!
!!  \exception  ALLOC Failed to allocate memory for node data
!!  \exception  SZE   Specified number of nodes is invalid
!!
!!  This routine allocates memory and initializes the state variable that
!!  contains the set of nodes for the problem.
! ------------------------------------------------------------------------
SUBROUTINE fld_init_node_exc_ (nnod, exc)
  INTEGER, INTENT(IN) :: nnod
  INTEGER, INTENT(OUT) :: exc
  INTEGER :: e        !> status code for memory allocation
  INTEGER :: i, j     !> loop variables

  !> make sure number of nodes is valid
  IF (nnod.LT.1 .OR. nnod.GT.MAX_NODES) THEN
    exc=SZE
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> only reallocate if new dimensions do not match existing dimensions
  IF (fld_numNode().NE.nnod) THEN

    !> ensure state variable is clear
    CALL fld_cleanNode()

    !> allocate memory for node data state variable
    ALLOCATE(nodes(nnod), STAT=e)

    !> make sure memory allocation was successful
    IF (e.NE.0) THEN
      exc=ALLOC
      CALL log_printLogMsg(exc,sdr)
      RETURN
    ELSE
      exc=OK
    END IF

  END IF

  !> initialize state variable

  ndof = 0

  DO i = 1,nnod

    !> node number
    nodes(i)%num   = i

    !> vector variables
    DO j = 1,NDIM
      nodes(i)%pt   %x(j)   = 0.d0
      nodes(i)%fix  %fix(j) = .FALSE.
      nodes(i)%dof  %dof(j) = 0
      nodes(i)%disp %u(j)   = 0.d0
      nodes(i)%vel  %v(j)   = 0.d0
      nodes(i)%acc  %a(j)   = 0.d0
      nodes(i)%body %a(j)   = 0.d0
    END DO

    !> tensor variables
    DO j = 1,NTNS
      nodes(i)%stress%sig(j) = 0.d0
      nodes(i)%strain%eps(j) = 0.d0
    END DO

  END DO

END SUBROUTINE fld_init_node_exc_


! ------------------------------------------------------------------------
!>  \brief Destructor for node data
!!
!!  This routine clears the memory allocated to the state variable that
!!  contains the set of node data for the problem. It also resets the
!!  degree of freedom counter (since there cannot be degrees of freedom
!!  without nodes).
! ------------------------------------------------------------------------
SUBROUTINE fld_clean_node_ ()

  IF (ALLOCATED(nodes)) DEALLOCATE(nodes)

  ndof = 0

END SUBROUTINE fld_clean_node_


! ------------------------------------------------------------------------
!>  \brief Getter for number of nodes
!!
!!  \return nnod  Number of nodes
!!
!!  This routine determines the number of data entries that have been
!!  allocated for node data. It does not check that the node data
!!  has been populated (i.e. changed from initial zero values).
! ------------------------------------------------------------------------
FUNCTION fld_num_node_ () RESULT(nnod)
  INTEGER :: nnod

  !> if data is initialized, return number of nodes
  IF (ALLOCATED(nodes)) THEN
    nnod = SIZE(nodes)
  ELSE
    nnod = 0   !> if not initialized, there are no nodes
  END IF

END FUNCTION fld_num_node_


! ------------------------------------------------------------------------
!>  \brief Getter for node coordinates (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!
!!  \return x     Coordinate value
!!
!!  This routine determines the value of the coordinate for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
FUNCTION fld_get_coord_ (i,j) RESULT(x)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION :: x

  x = nodes(i)%pt%x(j)

END FUNCTION fld_get_coord_


! ------------------------------------------------------------------------
!>  \brief Getter for node coordinates (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  exc   Error code
!!
!!  \return x     Coordinate value
!!
!!  \exception  POSIT The node number is not in [1..fld_numNode()] or the
!!                    coordinate index is not in [1..NDIM]
!!
!!  This routine determines the value of the coordinate for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
FUNCTION fld_get_coord_exc_ (i,j, exc) RESULT(x)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: x

  !> check that the node number is within the bounds of the nodes list
  !!  and that the coordinate index is within [1..NDIM]
  IF ( i.LT.1 .OR. i.GT.fld_numNode() &
        .OR. j.LT.1 .OR. j.GT.NDIM ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    x = 0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  x = fld_getCoord(i,j)

END FUNCTION fld_get_coord_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for node coordinates (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  x     Coordinate value
!!
!!  This routine sets the value of the coordinate for node number i in
!!  coordinate direction j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_coord_ (i,j, x)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: x

  nodes(i)%pt%x(j) = x

END SUBROUTINE fld_set_coord_


! ------------------------------------------------------------------------
!>  \brief Setter for node coordinates (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  x     Coordinate value
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  The value of the coordinate exceeds prescribed
!!                      limits
!!  \exception  POSIT   The node number is not in [1..fld_numNode()] or
!!                      the coordinate index is not in [1..NDIM]
!!
!!  This routine sets the value of the coordinate for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_coord_exc_ (i,j, x, exc)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: x
  INTEGER, INTENT(OUT) :: exc

  !> check that the node number is within the bounds of the nodes list
  !!  and that the coordinate index is within [1..NDIM]
  IF ( i.LT.1 .OR. i.GT.fld_numNode() &
        .OR. j.LT.1 .OR. j.GT.NDIM ) THEN

    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE IF (x.LT.COORD_MIN .OR. x.GT.COORD_MAX) THEN

    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL fld_setCoord(i,j, x)

END SUBROUTINE fld_set_coord_exc_


! ------------------------------------------------------------------------
!>  \brief Getter for fixities (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!
!!  \return fix   Fixity value
!!
!!  This routine determines the value of the fixity for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
FUNCTION fld_get_fix_ (i,j) RESULT(fix)
  INTEGER, INTENT(IN) :: i,j
  LOGICAL :: fix

  fix = nodes(i)%fix%fix(j)

END FUNCTION fld_get_fix_


! ------------------------------------------------------------------------
!>  \brief Getter for fixities (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  exc   Error code
!!
!!  \return fix   Fixity value
!!
!!  \exception  POSIT The node number is not in [1..fld_numNode()] or the
!!                    coordinate index is not in [1..NDIM]
!!
!!  This routine determines the value of the fixity for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
FUNCTION fld_get_fix_exc_ (i,j, exc) RESULT(fix)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: exc
  LOGICAL :: fix

  !> check that the node number is within the bounds of the nodes list
  !!  and that the coordinate index is within [1..NDIM]
  IF ( i.LT.1 .OR. i.GT.fld_numNode() &
        .OR. j.LT.1 .OR. j.GT.NDIM ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    fix = .FALSE.
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  fix = fld_getFix(i,j)

END FUNCTION fld_get_fix_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for fixities (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  fix   Fixity value
!!
!!  This routine sets the value of the fixity for node number i in
!!  coordinate direction j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_fix_ (i,j, fix)
  INTEGER, INTENT(IN) :: i,j
  LOGICAL, INTENT(IN) :: fix

  nodes(i)%fix%fix(j) = fix

END SUBROUTINE fld_set_fix_


! ------------------------------------------------------------------------
!>  \brief Setter for fixities (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  fix   Fixity value
!!  \param  exc   Error code
!!
!!  \exception  POSIT   The node number is not in [1..fld_numNode()] or
!!                      the coordinate index is not in [1..NDIM]
!!
!!  This routine sets the value of the coordinate for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_fix_exc_ (i,j, fix, exc)
  INTEGER, INTENT(IN) :: i,j
  LOGICAL, INTENT(IN) :: fix
  INTEGER, INTENT(OUT) :: exc

  !> check that the node number is within the bounds of the nodes list
  !!  and that the coordinate index is within [1..NDIM]
  IF ( i.LT.1 .OR. i.GT.fld_numNode() &
        .OR. j.LT.1 .OR. j.GT.NDIM ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL fld_setFix(i,j, fix)

END SUBROUTINE fld_set_fix_exc_


! ------------------------------------------------------------------------
!>  \brief Initializer for nodal degrees of freedom
!!
!!  \param  exc   Error code
!!
!!  \exception  SZE   The number of degrees of freedom exceeds the maximum
!!                    defined in the System Constants module
! ------------------------------------------------------------------------
SUBROUTINE fld_init_dof_exc_ (exc)
  INTEGER, INTENT(OUT), OPTIONAL :: exc
  INTEGER :: i,j    !> loop variables

  !> initialize degree of freedom counter
  ndof = 0

  !> loop through nodes, setting dof number for unfixed dofs
  DO i = 1,fld_numNode()
    DO j = 1,NDIM

      !> if dof is fixed, set dof number to 0
      IF ( fld_getFix(i,j) ) THEN

        nodes(i)%dof%dof(j) = 0

      !> otherwise, increment dof counter and set dof number
      ELSE

        ndof = ndof+1

        !> make sure number of dofs does not exceed limit
        IF (PRESENT(exc) .AND. ndof.GT.MAX_DOFS) THEN
          exc=SZE
          CALL log_printLogMsg(exc,sdr)
          ndof = 0
          RETURN
        END IF

        nodes(i)%dof%dof(j) = ndof

      END IF

    END DO
  END DO

  !> if the loop completed without returning, no exception was raised
  IF (PRESENT(exc)) exc=OK

END SUBROUTINE fld_init_dof_exc_


! ------------------------------------------------------------------------
!>  \brief Getter for number of degrees of freedom
!!
!!  \return n  Number of degrees of freedom
!!
!!  This routine determines the number of degrees of freedom for the
!!  system.
! ------------------------------------------------------------------------
FUNCTION fld_num_dof_ () RESULT(n)
  INTEGER :: n

  n = ndof

END FUNCTION fld_num_dof_


! ------------------------------------------------------------------------
!>  \brief Getter for degree of freedom (non-exception checking)
!!
!!  \param  i   Node number
!!  \param  j   Coordinate index
!!
!!  \return dof Degree of freedom number
!!
!!  This routine determines the index number associated with degree of
!!  freedom j for node i. Note that a return value of 0 indicates that the
!!  degree of freedom is fixed (and will not be necessary for the solution
!!  of the system).
! ------------------------------------------------------------------------
FUNCTION fld_get_dof_ (i,j) RESULT(dof)
  INTEGER, INTENT(IN) :: i,j
  INTEGER :: dof

  dof = nodes(i)%dof%dof(j)

END FUNCTION fld_get_dof_


! ------------------------------------------------------------------------
!>  \brief Getter for degree of freedom (exception checking)
!!
!!  \param  i   Node number
!!  \param  j   Coordinate index
!!  \param  exc Error code
!!
!!  \return dof Degree of freedom number
!!
!!  \exception  POSIT The node number is not in [1..fld_numNode()] or the
!!                    coordinate index is not in [1..NDIM]
!!
!!  This routine determines the index number associated with degree of
!!  freedom j for node i. Note that a return value of 0 indicates that the
!!  degree of freedom is fixed (and will not be necessary for the solution
!!  of the system).
! ------------------------------------------------------------------------
FUNCTION fld_get_dof_exc_ (i,j, exc) RESULT(dof)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: exc
  INTEGER :: dof

  !> check that the given indices are valid
  IF (  i.LT.1 .OR. i.GT.fld_numNode()  &
        .OR. j.LT.1 .OR. j.GT.NDIM )    THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    dof=0
    RETURN
  ELSE
    exc=OK
  END IF

  !> make sure dofs have been initialized
  IF (ndof.EQ.0) THEN
    dof=0
  ELSE
    !> call non-exception version
    dof = fld_getDof(i,j)
  END IF

END FUNCTION fld_get_dof_exc_


! ------------------------------------------------------------------------
!>  \brief Getter for node displacements (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!
!!  \return u     Displacement value
!!
!!  This routine determines the value of the displacement for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
FUNCTION fld_get_disp_ (i,j) RESULT(u)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION :: u

  u = nodes(i)%disp%u(j)

END FUNCTION fld_get_disp_


! ------------------------------------------------------------------------
!>  \brief Getter for node displacements (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  exc   Error code
!!
!!  \return u     Displacement value
!!
!!  \exception  POSIT The node number is not in [1..fld_numNode()] or the
!!                    coordinate index is not in [1..NDIM]
!!
!!  This routine determines the value of the displacement for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
FUNCTION fld_get_disp_exc_ (i,j, exc) RESULT(u)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: u

  !> check that the node number is within the bounds of the nodes list
  !!  and that the coordinate index is within [1..NDIM]
  IF ( i.LT.1 .OR. i.GT.fld_numNode() &
        .OR. j.LT.1 .OR. j.GT.NDIM ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    u = 0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  u = fld_getDisp(i,j)

END FUNCTION fld_get_disp_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for node displacements (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  u     Displacement value
!!
!!  This routine sets the value of the displacement for node number i in
!!  coordinate direction j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_disp_ (i,j, u)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: u

  nodes(i)%disp%u(j) = u

END SUBROUTINE fld_set_disp_


! ------------------------------------------------------------------------
!>  \brief Setter for node displacements (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  u     Displacement value
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  The value of the displacement exceeds prescribed
!!                      limits
!!  \exception  POSIT   The node number is not in [1..fld_numNode()] or
!!                      the coordinate index is not in [1..NDIM]
!!
!!  This routine sets the value of the displacement for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_disp_exc_ (i,j, u, exc)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: u
  INTEGER, INTENT(OUT) :: exc

  !> check that the node number is within the bounds of the nodes list
  !!  and that the coordinate index is within [1..NDIM]
  IF ( i.LT.1 .OR. i.GT.fld_numNode() &
        .OR. j.LT.1 .OR. j.GT.NDIM ) THEN

    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE IF (u.LT.DISP_MIN .OR. u.GT.DISP_MAX) THEN

    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL fld_setDisp(i,j, u)

END SUBROUTINE fld_set_disp_exc_


! ------------------------------------------------------------------------
!>  \brief Getter for node velocities (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!
!!  \return v     Velocity value
!!
!!  This routine determines the value of the velocity for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
FUNCTION fld_get_vel_ (i,j) RESULT(v)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION :: v

  v = nodes(i)%vel%v(j)

END FUNCTION fld_get_vel_


! ------------------------------------------------------------------------
!>  \brief Getter for node velocities (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  exc   Error code
!!
!!  \return v     Velocity value
!!
!!  \exception  POSIT The node number is not in [1..fld_numNode()] or the
!!                    coordinate index is not in [1..NDIM]
!!
!!  This routine determines the value of the velocity for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
FUNCTION fld_get_vel_exc_ (i,j, exc) RESULT(v)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: v

  !> check that the node number is within the bounds of the nodes list
  !!  and that the coordinate index is within [1..NDIM]
  IF ( i.LT.1 .OR. i.GT.fld_numNode() &
        .OR. j.LT.1 .OR. j.GT.NDIM ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    v = 0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  v = fld_getVel(i,j)

END FUNCTION fld_get_vel_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for node velocities (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  v     Velocity value
!!
!!  This routine sets the value of the velocity for node number i in
!!  coordinate direction j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_vel_ (i,j, v)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: v

  nodes(i)%vel%v(j) = v

END SUBROUTINE fld_set_vel_


! ------------------------------------------------------------------------
!>  \brief Setter for node velocities (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  v     Velocity value
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  The value of the velocity exceeds prescribed
!!                      limits
!!  \exception  POSIT   The node number is not in [1..fld_numNode()] or
!!                      the coordinate index is not in [1..NDIM]
!!
!!  This routine sets the value of the velocity for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_vel_exc_ (i,j, v, exc)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: v
  INTEGER, INTENT(OUT) :: exc

  !> check that the node number is within the bounds of the nodes list
  !!  and that the coordinate index is within [1..NDIM]
  IF ( i.LT.1 .OR. i.GT.fld_numNode() &
        .OR. j.LT.1 .OR. j.GT.NDIM ) THEN

    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE IF (v.LT.VEL_MIN .OR. v.GT.VEL_MAX) THEN

    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL fld_setVel(i,j, v)

END SUBROUTINE fld_set_vel_exc_


! ------------------------------------------------------------------------
!>  \brief Getter for node accelerations (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!
!!  \return a     Acceleration value
!!
!!  This routine determines the value of the acceleration for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
FUNCTION fld_get_acc_ (i,j) RESULT(a)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION :: a

  a = nodes(i)%acc%a(j)

END FUNCTION fld_get_acc_


! ------------------------------------------------------------------------
!>  \brief Getter for node accelerations (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  exc   Error code
!!
!!  \return a     Acceleration value
!!
!!  \exception  POSIT The node number is not in [1..fld_numNode()] or the
!!                    coordinate index is not in [1..NDIM]
!!
!!  This routine determines the value of the acceleration for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
FUNCTION fld_get_acc_exc_ (i,j, exc) RESULT(a)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: a

  !> check that the node number is within the bounds of the nodes list
  !!  and that the coordinate index is within [1..NDIM]
  IF ( i.LT.1 .OR. i.GT.fld_numNode() &
        .OR. j.LT.1 .OR. j.GT.NDIM ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    a = 0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  a = fld_getAcc(i,j)

END FUNCTION fld_get_acc_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for node accelerations (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  a     Acceleration value
!!
!!  This routine sets the value of the acceleration for node number i in
!!  coordinate direction j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_acc_ (i,j, a)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: a

  nodes(i)%acc%a(j) = a

END SUBROUTINE fld_set_acc_


! ------------------------------------------------------------------------
!>  \brief Setter for node accelerations (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  a     Acceleration value
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  The value of the acceleration exceeds prescribed
!!                      limits
!!  \exception  POSIT   The node number is not in [1..fld_numNode()] or
!!                      the coordinate index is not in [1..NDIM]
!!
!!  This routine sets the value of the acceleration for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_acc_exc_ (i,j, a, exc)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: a
  INTEGER, INTENT(OUT) :: exc

  !> check that the node number is within the bounds of the nodes list
  !!  and that the coordinate index is within [1..NDIM]
  IF ( i.LT.1 .OR. i.GT.fld_numNode() &
        .OR. j.LT.1 .OR. j.GT.NDIM ) THEN

    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE IF (a.LT.ACC_MIN .OR. a.GT.ACC_MAX) THEN

    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL fld_setAcc(i,j, a)

END SUBROUTINE fld_set_acc_exc_


! ------------------------------------------------------------------------
!>  \brief Getter for node body accelerations (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!
!!  \return a     Body acceleration value
!!
!!  This routine determines the value of the body acceleration for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
FUNCTION fld_get_body_acc_ (i,j) RESULT(a)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION :: a

  a = nodes(i)%body%a(j)

END FUNCTION fld_get_body_acc_


! ------------------------------------------------------------------------
!>  \brief Getter for node body accelerations (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  exc   Error code
!!
!!  \return a     Body acceleration value
!!
!!  \exception  POSIT The node number is not in [1..fld_numNode()] or the
!!                    coordinate index is not in [1..NDIM]
!!
!!  This routine determines the value of the body acceleration for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
FUNCTION fld_get_body_acc_exc_ (i,j, exc) RESULT(a)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: a

  !> check that the node number is within the bounds of the nodes list
  !!  and that the coordinate index is within [1..NDIM]
  IF ( i.LT.1 .OR. i.GT.fld_numNode() &
        .OR. j.LT.1 .OR. j.GT.NDIM ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    a = 0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  a = fld_getBodyAcc(i,j)

END FUNCTION fld_get_body_acc_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for node body accelerations (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  a     Body acceleration value
!!
!!  This routine sets the value of the body acceleration for node number i
!!  in coordinate direction j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_body_acc_ (i,j, a)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: a

  nodes(i)%body%a(j) = a

END SUBROUTINE fld_set_body_acc_


! ------------------------------------------------------------------------
!>  \brief Setter for node body accelerations (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Coordinate index
!!  \param  a     Body acceleration value
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  The value of the body acceleration exceeds
!!                      prescribed limits
!!  \exception  POSIT   The node number is not in [1..fld_numNode()] or
!!                      the coordinate index is not in [1..NDIM]
!!
!!  This routine sets the value of the body acceleration for node
!!  number i in coordinate direction j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_body_acc_exc_ (i,j, a, exc)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: a
  INTEGER, INTENT(OUT) :: exc

  !> check that the node number is within the bounds of the nodes list
  !!  and that the coordinate index is within [1..NDIM]
  IF ( i.LT.1 .OR. i.GT.fld_numNode() &
        .OR. j.LT.1 .OR. j.GT.NDIM ) THEN

    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE IF (a.LT.ACC_MIN .OR. a.GT.ACC_MAX) THEN

    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL fld_setBodyAcc(i,j, a)

END SUBROUTINE fld_set_body_acc_exc_


! ------------------------------------------------------------------------
!>  \brief Getter for node stresses (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Tensor index
!!
!!  \return s     Stress value
!!
!!  This routine determines the value of the stress for node
!!  number i for tensor index j.
! ------------------------------------------------------------------------
FUNCTION fld_get_stress_node_ (i,j) RESULT(s)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION :: s

  s = nodes(i)%stress%sig(j)

END FUNCTION fld_get_stress_node_


! ------------------------------------------------------------------------
!>  \brief Getter for node stresses (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Tensor index
!!  \param  exc   Error code
!!
!!  \return s     Stress value
!!
!!  \exception  POSIT The node number is not in [1..fld_numNode()] or the
!!                    tensor index is not in [1..NTNS]
!!
!!  This routine determines the value of the stress for node
!!  number i for tensor index j.
! ------------------------------------------------------------------------
FUNCTION fld_get_stress_node_exc_ (i,j, exc) RESULT(s)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: s

  !> check that the node number is within the bounds of the nodes list
  !!  and that the tensor index is within [1..NTNS]
  IF ( i.LT.1 .OR. i.GT.fld_numNode() &
        .OR. j.LT.1 .OR. j.GT.NTNS ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    s = 0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  s = fld_getStressNode(i,j)

END FUNCTION fld_get_stress_node_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for node stresses (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Tensor index
!!  \param  s     Stress value
!!
!!  This routine sets the value of the stress for node number i
!!  for tensor index j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_stress_node_ (i,j, s)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: s

  nodes(i)%stress%sig(j) = s

END SUBROUTINE fld_set_stress_node_


! ------------------------------------------------------------------------
!>  \brief Setter for node stresses (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Tensor index
!!  \param  s     Stress value
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  The value of the stress exceeds prescribed limits
!!  \exception  POSIT   The node number is not in [1..fld_numNode()] or
!!                      the tensor index is not in [1..NTNS]
!!
!!  This routine sets the value of the stress for node number i for
!!  tensor index j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_stress_node_exc_ (i,j, s, exc)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: s
  INTEGER, INTENT(OUT) :: exc

  !> check that the node number is within the bounds of the nodes list
  !!  and that the tensor index is within [1..NTNS]
  IF ( i.LT.1 .OR. i.GT.fld_numNode() &
        .OR. j.LT.1 .OR. j.GT.NTNS ) THEN

    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE IF (s.LT.SIG_MIN .OR. s.GT.SIG_MAX) THEN

    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL fld_setStressNode(i,j, s)

END SUBROUTINE fld_set_stress_node_exc_


! ------------------------------------------------------------------------
!>  \brief Getter for node strains (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Tensor index
!!
!!  \return s     Strain value
!!
!!  This routine determines the value of the strain for node
!!  number i for tensor index j.
! ------------------------------------------------------------------------
FUNCTION fld_get_strain_node_ (i,j) RESULT(s)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION :: s

  s = nodes(i)%strain%eps(j)

END FUNCTION fld_get_strain_node_


! ------------------------------------------------------------------------
!>  \brief Getter for node strains (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Tensor index
!!  \param  exc   Error code
!!
!!  \return s     Strain value
!!
!!  \exception  POSIT The node number is not in [1..fld_numNode()] or the
!!                    tensor index is not in [1..NTNS]
!!
!!  This routine determines the value of the strain for node
!!  number i for tensor index j.
! ------------------------------------------------------------------------
FUNCTION fld_get_strain_node_exc_ (i,j, exc) RESULT(s)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: s

  !> check that the node number is within the bounds of the nodes list
  !!  and that the tensor index is within [1..NTNS]
  IF ( i.LT.1 .OR. i.GT.fld_numNode() &
        .OR. j.LT.1 .OR. j.GT.NTNS ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    s = 0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  s = fld_getStrainNode(i,j)

END FUNCTION fld_get_strain_node_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for node strains (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Tensor index
!!  \param  s     Strain value
!!
!!  This routine sets the value of the strain for node number i
!!  for tensor index j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_strain_node_ (i,j, s)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: s

  nodes(i)%strain%eps(j) = s

END SUBROUTINE fld_set_strain_node_


! ------------------------------------------------------------------------
!>  \brief Setter for node strains (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Tensor index
!!  \param  s     Strain value
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  The value of the strain exceeds prescribed limits
!!  \exception  POSIT   The node number is not in [1..fld_numNode()] or
!!                      the tensor index is not in [1..NTNS]
!!
!!  This routine sets the value of the strain for node number i for
!!  tensor index j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_strain_node_exc_ (i,j, s, exc)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: s
  INTEGER, INTENT(OUT) :: exc

  !> check that the node number is within the bounds of the nodes list
  !!  and that the tensor index is within [1..NTNS]
  IF ( i.LT.1 .OR. i.GT.fld_numNode() &
        .OR. j.LT.1 .OR. j.GT.NTNS ) THEN

    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE IF (s.LT.STR_MIN .OR. s.GT.STR_MAX) THEN

    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL fld_setStrainNode(i,j, s)

END SUBROUTINE fld_set_strain_node_exc_


! ------------------------------------------------------------------------
!>  \brief Constructor for element data (non-exception checking)
!!
!!  \param  nel  Number of elements
!!
!!  This routine allocates memory and initializes the state variable that
!!  contains the set of elements for the problem.
! ------------------------------------------------------------------------
SUBROUTINE fld_init_elem_ (nel)
  INTEGER, INTENT(IN) :: nel
  INTEGER :: i, j     !> loop variables

  !> only reallocate if new dimensions do not match existing dimensions
  IF (fld_numElem().NE.nel) THEN

    !> ensure state variable is clear
    CALL fld_cleanElem()

    !> allocate memory for element data state variable
    ALLOCATE(elements(nel))

  END IF

  !> initialize state variable
  DO i = 1,nel

    !> element number
    elements(i)%num   = i

    !> material number
    elements(i)%mtl = 0

    !> connectivity
    DO j = 1,NNODEL
      elements(i)%ico%nd(j)   = 0
    END DO

    !> tensor variables
    DO j = 1,NTNS
      elements(i)%stress%sig(j) = 0.d0
      elements(i)%strain%eps(j) = 0.d0
    END DO

  END DO

END SUBROUTINE fld_init_elem_


! ------------------------------------------------------------------------
!>  \brief Constructor for element data (exception checking)
!!
!!  \param  nel   Number of elements
!!  \param  exc   Error code
!!
!!  \exception  ALLOC Failed to allocate memory for element data
!!  \exception  SZE   Specified number of elements is invalid
!!
!!  This routine allocates memory and initializes the state variable that
!!  contains the set of elements for the problem.
! ------------------------------------------------------------------------
SUBROUTINE fld_init_elem_exc_ (nel, exc)
  INTEGER, INTENT(IN) :: nel
  INTEGER, INTENT(OUT) :: exc
  INTEGER :: e        !> status code for memory allocation
  INTEGER :: i, j     !> loop variables

  !> make sure number of elements is valid
  IF (nel.LT.1 .OR. nel.GT.MAX_ELEMENTS) THEN
    exc=SZE
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> only reallocate if new dimensions do not match existing dimensions
  IF (fld_numElem().NE.nel) THEN

    !> ensure state variable is clear
    CALL fld_cleanElem()

    !> allocate memory for element data state variable
    ALLOCATE(elements(nel), STAT=e)

    !> make sure memory allocation was successful
    IF (e.NE.0) THEN
      exc=ALLOC
      CALL log_printLogMsg(exc,sdr)
      RETURN
    ELSE
      exc=OK
    END IF

  END IF

  !> initialize state variable
  DO i = 1,nel

    !> element number
    elements(i)%num   = i

    !> material number
    elements(i)%mtl = 0

    !> connectivity
    DO j = 1,NNODEL
      elements(i)%ico%nd(j)   = 0
    END DO

    !> tensor variables
    DO j = 1,NTNS
      elements(i)%stress%sig(j) = 0.d0
      elements(i)%strain%eps(j) = 0.d0
    END DO

  END DO

END SUBROUTINE fld_init_elem_exc_


! ------------------------------------------------------------------------
!>  \brief Destructor for element data
!!
!!  This routine clears the memory allocated to the state variable that
!!  contains the set of element data for the problem.
! ------------------------------------------------------------------------
SUBROUTINE fld_clean_elem_ ()

  IF (ALLOCATED(elements)) DEALLOCATE(elements)

END SUBROUTINE fld_clean_elem_


! ------------------------------------------------------------------------
!>  \brief Getter for number of elements
!!
!!  \return nel  Number of elements
!!
!!  This routine determines the number of data entries that have been
!!  allocated for element data. It does not check that the element data
!!  has been populated (i.e. changed from initial zero values).
! ------------------------------------------------------------------------
FUNCTION fld_num_elem_ () RESULT(nel)
  INTEGER :: nel

  !> if data is initialized, return number of elements
  IF (ALLOCATED(elements)) THEN
    nel = SIZE(elements)
  ELSE
    nel = 0   !> if not initialized, there are no elements
  END IF

END FUNCTION fld_num_elem_


! ------------------------------------------------------------------------
!>  \brief Getter for connectivity (non-exception checking)
!!
!!  \param  i     Element number
!!  \param  j     Local node index
!!
!!  \return n     Global node number
!!
!!  This routine determines the global node number associated with local
!!  node j for element i.
! ------------------------------------------------------------------------
FUNCTION fld_get_connect_ (i,j) RESULT(n)
  INTEGER, INTENT(IN) :: i,j
  INTEGER :: n

  n = elements(i)%ico%nd(j)

END FUNCTION fld_get_connect_


! ------------------------------------------------------------------------
!>  \brief Getter for connectivity (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Local node index
!!  \param  exc   Error code
!!
!!  \return n     Global node number
!!
!!  \exception  POSIT The element number is not in [1..fld_numElem()] or
!!                    the local node index is not in [1..NNODEL]
!!
!!  This routine determines the global node number associated with local
!!  node j for element i.
! ------------------------------------------------------------------------
FUNCTION fld_get_connect_exc_ (i,j, exc) RESULT(n)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: exc
  INTEGER :: n

  !> check that the element number is within the bounds of the elements
  !! list and that the local node index is within [1..NNODEL]
  IF ( i.LT.1 .OR. i.GT.fld_numElem() &
        .OR. j.LT.1 .OR. j.GT.NNODEL ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    n = 0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  n = fld_getConnect(i,j)

END FUNCTION fld_get_connect_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for connectivity (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Local node index
!!  \param  n     Global node number
!!
!!  This routine sets the global node number associated with local
!!  node j for element i.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_connect_ (i,j, n)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(IN) :: n

  elements(i)%ico%nd(j) = n

END SUBROUTINE fld_set_connect_


! ------------------------------------------------------------------------
!>  \brief Setter for connectivity (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Local node index
!!  \param  n     Global node number
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  The value of the node number exceeds the number of
!!                      nodes
!!  \exception  POSIT   The element number is not in [1..fld_numElem()] or
!!                      the local node index is not in [1..NNODEL]
!!
!!  This routine sets the global node number associated with local
!!  node j for element i.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_connect_exc_ (i,j, n, exc)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(IN) :: n
  INTEGER, INTENT(OUT) :: exc

  !> check that the element number is within the bounds of the elements
  !! list and that the local node index is within [1..NNODEL]
  IF ( i.LT.1 .OR. i.GT.fld_numElem() &
        .OR. j.LT.1 .OR. j.GT.NNODEL ) THEN

    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE IF (n.LT.1 .OR. n.GT.fld_numNode()) THEN

    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL fld_setConnect(i,j, n)

END SUBROUTINE fld_set_connect_exc_


! ------------------------------------------------------------------------
!>  \brief Compute the volume of an element (non-exception checking)
!!
!!  \param  i     Element number
!!
!!  \return vol   Volume of element
!!
!!  This routine computes the volume of a body element. Note that for a
!!  2-D plane strain element this is equivalent to the area of the element
!!  (considering a unit thickness). This is computed as the following
!!  determinant:
!!
!!    V   =  0.5 *  | x1  y1  1 |
!!                  | x2  y2  1 |
!!                  | x3  y3  1 |
!!
!!  where xj and yj are the (x,y) coordinates of node j of element i.
! ------------------------------------------------------------------------
FUNCTION fld_vol_elem_ (i) RESULT(vol)
  INTEGER, INTENT(IN) :: i
  DOUBLE PRECISION :: vol
  DOUBLE PRECISION, DIMENSION(NNODEL,NDIM) :: x  !> element coords
  INTEGER :: j,k    !> loop variable

  !> get element coords
  DO k = 1,NDIM
    DO j = 1,NNODEL
      x(j,k) = fld_getCoord(elements(i)%ico%nd(j), k)
    END DO
  END DO

  !> compute volume
  vol = ( x(1,1)*x(2,2) - x(2,1)*x(1,2) )   &
        - ( x(1,1)*x(3,2) - x(3,1)*x(1,2) ) &
        + ( x(2,1)*x(3,2) - x(3,1)*x(2,2) )

  vol = 0.5d0 * vol

END FUNCTION fld_vol_elem_


! ------------------------------------------------------------------------
!>  \brief Compute the volume of an element (exception checking)
!!
!!  \param  i     Element number
!!  \param  exc   Error code
!!
!!  \return vol   Volume of element
!!
!!  \exception  POSIT   The specified element number is not within the
!!                      range [1..fld_numElem()]
!!
!!  This routine computes the volume of a body element. Note that for a
!!  2-D plane strain element this is equivalent to the area of the element
!!  (considering a unit thickness). This is computed as the following
!!  determinant:
!!
!!    V   =  0.5 *  | x1  y1  1 |
!!                  | x2  y2  1 |
!!                  | x3  y3  1 |
!!
!!  where xj and yj are the (x,y) coordinates of node j of element i.
! ------------------------------------------------------------------------
FUNCTION fld_vol_elem_exc_ (i, exc) RESULT(vol)
  INTEGER, INTENT(IN) :: i
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: vol

  !> make sure the element number is valid
  IF (i.LT.1 .OR. i.GT.fld_numElem()) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    vol = 0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  vol = fld_volElem(i)

END FUNCTION fld_vol_elem_exc_


! ------------------------------------------------------------------------
!>  \brief Getter for material (non-exception checking)
!!
!!  \param  i     Element number
!!
!!  \return m     Material number
!!
!!  This routine determines the material number associated with element i
! ------------------------------------------------------------------------
FUNCTION fld_get_material_ (i) RESULT(m)
  INTEGER, INTENT(IN) :: i
  INTEGER :: m

  m = elements(i)%mtl

END FUNCTION fld_get_material_


! ------------------------------------------------------------------------
!>  \brief Getter for material (exception checking)
!!
!!  \param  i     Element number
!!  \param  exc   Error code
!!
!!  \return m     Material number
!!
!!  \exception  POSIT The element number is not in [1..fld_numElem()]
!!
!!  This routine determines the material number associated with element i
! ------------------------------------------------------------------------
FUNCTION fld_get_material_exc_ (i, exc) RESULT(m)
  INTEGER, INTENT(IN) :: i
  INTEGER, INTENT(OUT) :: exc
  INTEGER :: m

  !> check that the element number is within the bounds of the elements
  !! list
  IF ( i.LT.1 .OR. i.GT.fld_numElem() ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    m = 0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  m = fld_getMaterial(i)

END FUNCTION fld_get_material_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for material (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  m     Material number
!!
!!  This routine sets the material number associated with element i.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_material_ (i, m)
  INTEGER, INTENT(IN) :: i
  INTEGER, INTENT(IN) :: m

  elements(i)%mtl = m

END SUBROUTINE fld_set_material_


! ------------------------------------------------------------------------
!>  \brief Setter for material (exception checking)
!!
!!  \param  i     Element number
!!  \param  n     Material number
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  The value of the material number exceeds the
!!                      number of materials
!!  \exception  POSIT   The element number is not in [1..fld_numElem()]
!!
!!  This routine sets the material number associated with element i.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_material_exc_ (i, m, exc)
  INTEGER, INTENT(IN) :: i
  INTEGER, INTENT(IN) :: m
  INTEGER, INTENT(OUT) :: exc

  !> check that the element number is within the bounds of the elements
  !! list
  IF ( i.LT.1 .OR. i.GT.fld_numElem() ) THEN

    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE IF (m.LT.1 .OR. m.GT.mtl_numMtl()) THEN

    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL fld_setMaterial(i, m)

END SUBROUTINE fld_set_material_exc_


! ------------------------------------------------------------------------
!>  \brief Getter for element stresses (non-exception checking)
!!
!!  \param  i     Element number
!!  \param  j     Tensor index
!!
!!  \return s     Stress value
!!
!!  This routine determines the value of the stress for element
!!  number i for tensor index j.
! ------------------------------------------------------------------------
FUNCTION fld_get_stress_elem_ (i,j) RESULT(s)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION :: s

  s = elements(i)%stress%sig(j)

END FUNCTION fld_get_stress_elem_


! ------------------------------------------------------------------------
!>  \brief Getter for element stresses (exception checking)
!!
!!  \param  i     Element number
!!  \param  j     Tensor index
!!  \param  exc   Error code
!!
!!  \return s     Stress value
!!
!!  \exception  POSIT The element number is not in [1..fld_numElem()] or
!!                    the tensor index is not in [1..NTNS]
!!
!!  This routine determines the value of the stress for element
!!  number i for tensor index j.
! ------------------------------------------------------------------------
FUNCTION fld_get_stress_elem_exc_ (i,j, exc) RESULT(s)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: s

  !> check that the element number is within the bounds of the elements
  !! list and that the tensor index is within [1..NTNS]
  IF ( i.LT.1 .OR. i.GT.fld_numElem() &
        .OR. j.LT.1 .OR. j.GT.NTNS ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    s = 0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  s = fld_getStressElem(i,j)

END FUNCTION fld_get_stress_elem_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for element stresses (non-exception checking)
!!
!!  \param  i     Element number
!!  \param  j     Tensor index
!!  \param  s     Stress value
!!
!!  This routine sets the value of the stress for element number i
!!  for tensor index j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_stress_elem_ (i,j, s)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: s

  elements(i)%stress%sig(j) = s

END SUBROUTINE fld_set_stress_elem_


! ------------------------------------------------------------------------
!>  \brief Setter for element stresses (exception checking)
!!
!!  \param  i     Element number
!!  \param  j     Tensor index
!!  \param  s     Stress value
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  The value of the stress exceeds prescribed limits
!!  \exception  POSIT   The element number is not in [1..fld_numElem()] or
!!                      the tensor index is not in [1..NTNS]
!!
!!  This routine sets the value of the stress for node number i for
!!  tensor index j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_stress_elem_exc_ (i,j, s, exc)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: s
  INTEGER, INTENT(OUT) :: exc

  !> check that the element number is within the bounds of the elements
  !! list and that the tensor index is within [1..NTNS]
  IF ( i.LT.1 .OR. i.GT.fld_numElem() &
        .OR. j.LT.1 .OR. j.GT.NTNS ) THEN

    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE IF (s.LT.SIG_MIN .OR. s.GT.SIG_MAX) THEN

    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL fld_setStressElem(i,j, s)

END SUBROUTINE fld_set_stress_elem_exc_


! ------------------------------------------------------------------------
!>  \brief Getter for element strains (non-exception checking)
!!
!!  \param  i     Element number
!!  \param  j     Tensor index
!!
!!  \return s     Strain value
!!
!!  This routine determines the value of the strain for element
!!  number i for tensor index j.
! ------------------------------------------------------------------------
FUNCTION fld_get_strain_elem_ (i,j) RESULT(s)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION :: s

  s = elements(i)%strain%eps(j)

END FUNCTION fld_get_strain_elem_


! ------------------------------------------------------------------------
!>  \brief Getter for element strains (exception checking)
!!
!!  \param  i     Element number
!!  \param  j     Tensor index
!!  \param  exc   Error code
!!
!!  \return s     Strain value
!!
!!  \exception  POSIT The element number is not in [1..fld_numElem()] or
!!                    the tensor index is not in [1..NTNS]
!!
!!  This routine determines the value of the strain for element
!!  number i for tensor index j.
! ------------------------------------------------------------------------
FUNCTION fld_get_strain_elem_exc_ (i,j, exc) RESULT(s)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: s

  !> check that the element number is within the bounds of the elements
  !! list and that the tensor index is within [1..NTNS]
  IF ( i.LT.1 .OR. i.GT.fld_numElem() &
        .OR. j.LT.1 .OR. j.GT.NTNS ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    s = 0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  s = fld_getStrainElem(i,j)

END FUNCTION fld_get_strain_elem_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for element strains (non-exception checking)
!!
!!  \param  i     Element number
!!  \param  j     Tensor index
!!  \param  s     Strain value
!!
!!  This routine sets the value of the strain for element number i
!!  for tensor index j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_strain_elem_ (i,j, s)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: s

  elements(i)%strain%eps(j) = s

END SUBROUTINE fld_set_strain_elem_


! ------------------------------------------------------------------------
!>  \brief Setter for element strains (exception checking)
!!
!!  \param  i     Element number
!!  \param  j     Tensor index
!!  \param  s     Strain value
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  The value of the strain exceeds prescribed limits
!!  \exception  POSIT   The element number is not in [1..fld_numElem()] or
!!                      the tensor index is not in [1..NTNS]
!!
!!  This routine sets the value of the strain for element number i for
!!  tensor index j.
! ------------------------------------------------------------------------
SUBROUTINE fld_set_strain_elem_exc_ (i,j, s, exc)
  INTEGER, INTENT(IN) :: i,j
  DOUBLE PRECISION, INTENT(IN) :: s
  INTEGER, INTENT(OUT) :: exc

  !> check that the element number is within the bounds of the elements
  !! list and that the tensor index is within [1..NTNS]
  IF ( i.LT.1 .OR. i.GT.fld_numElem() &
        .OR. j.LT.1 .OR. j.GT.NTNS ) THEN

    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE IF (s.LT.STR_MIN .OR. s.GT.STR_MAX) THEN

    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL fld_setStrainElem(i,j, s)

END SUBROUTINE fld_set_strain_elem_exc_



! ************************************************************************
! LOCAL FUNCTIONS
! ************************************************************************

! none

END MODULE field_data
