! ------------------------------------------------------------------------
!>  \brief Module for Body Element Interpolation
! ------------------------------------------------------------------------
MODULE body_element_interpolation
USE system_constants    !> Global constants (for size of shape function matrix)
USE dense_matrix_def    !> Dense Matrix data type

IMPLICIT NONE
PRIVATE



! ************************************************************************
! EXPORTS
! ************************************************************************


!> Exported interfaces
PUBLIC :: bshp_shape



! ************************************************************************
! LOCAL CONSTANTS
! ************************************************************************

! none


! ************************************************************************
! DATA TYPES
! ************************************************************************

! none



! ************************************************************************
! INTERFACES
! ************************************************************************


!> \brief Interface to shape function access program
INTERFACE bshp_shape
  MODULE PROCEDURE bshp_shape_3_noded_tri_
END INTERFACE bshp_shape



CONTAINS



! ************************************************************************
! ACCESS PROGRAMS
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Shape function matrix for body element
!!
!!  \param  l1  Area coordinate 1
!!  \param  l2  Area coordinate 2
!!  \param  N   Shape function matrix
!!
!!  This routine builds the shape function matrix for a 3-noded triangular
!!  body element. It has the following form:
!!
!!    N = [ l1   0  l2   0  l3   0 ]
!!        [  0  l1   0  l2   0  l3 ]
!!
!!  where l3 = 1-l1-l2 and l1+l2 <= 1 is assumed.
! ------------------------------------------------------------------------
SUBROUTINE bshp_shape_3_noded_tri_ (l1,l2, N)
  DOUBLE PRECISION, INTENT(IN) :: l1,l2
  TYPE(matrixT), INTENT(INOUT) :: N
  DOUBLE PRECISION :: l3    !> Area coordinate 3

  !> compute third area coordinate
  l3 = 1.d0-l1-l2

  !> initialize shape function matrix
  CALL dm_init(N, NDIM, NDIM*NNODEL)

  !> build shape function matrix
  CALL dm_set(N, 1,1, l1)
  CALL dm_set(N, 2,2, l1)
  CALL dm_set(N, 1,3, l2)
  CALL dm_set(N, 2,4, l2)
  CALL dm_set(N, 1,5, l3)
  CALL dm_set(N, 2,6, l3)

END SUBROUTINE bshp_shape_3_noded_tri_



! ************************************************************************
! LOCAL FUNCTIONS
! ************************************************************************

! none

END MODULE body_element_interpolation




! ------------------------------------------------------------------------
!>  \brief Module for Traction Element Interpolation
! ------------------------------------------------------------------------
MODULE traction_element_interpolation
USE system_constants    !> Global constants (for size of shape function matrix)
USE dense_matrix_def    !> Dense Matrix data type
USE field_data          !> Field Data module
USE boundary_data          !> Boundary Data module

IMPLICIT NONE
PRIVATE



! ************************************************************************
! EXPORTS
! ************************************************************************


!> Exported interfaces
PUBLIC :: tshp_shape, tshp_transform



! ************************************************************************
! LOCAL CONSTANTS
! ************************************************************************

! none


! ************************************************************************
! DATA TYPES
! ************************************************************************

! none



! ************************************************************************
! INTERFACES
! ************************************************************************


!> \brief Interface to shape function access program
INTERFACE tshp_shape
  MODULE PROCEDURE tshp_shape_2_noded_linear_
END INTERFACE tshp_shape

!> \brief Interface to transformation matrix access program
INTERFACE tshp_transform
  MODULE PROCEDURE tshp_transform_2_noded_linear_
END INTERFACE tshp_transform



CONTAINS



! ************************************************************************
! ACCESS PROGRAMS
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Shape function matrix for traction element
!!
!!  \param  s   Local coordinate (0 <= s <= 1)
!!  \param  Nt  Shape function matrix
!!
!!  This routine builds the shape function matrix for a 2-noded traction
!!  element. It has the following form:
!!
!!    Nt = [ 1-s   0    s   0 ]
!!         [  0   1-s   0   s ]
! ------------------------------------------------------------------------
SUBROUTINE tshp_shape_2_noded_linear_ (s, Nt)
  DOUBLE PRECISION, INTENT(IN) :: s
  TYPE(matrixT), INTENT(INOUT) :: Nt
  DOUBLE PRECISION :: one_minus_s   !> for efficiency

  !> compute 1-s
  one_minus_s = 1.d0-s

  !> initialize shape function matrix
  CALL dm_init(Nt, NDIM, NDIM*NNODELB)

  !> build shape function matrix
  CALL dm_set(Nt, 1,1, one_minus_s)
  CALL dm_set(Nt, 2,2, one_minus_s)
  CALL dm_set(Nt, 1,3, s)
  CALL dm_set(Nt, 2,4, s)

END SUBROUTINE tshp_shape_2_noded_linear_



! ------------------------------------------------------------------------
!>  \brief Transformation matrix for traction element
!!
!!  \param  i   Element number
!!  \param  T   Transformation matrix
!!
!!  This routine builds the transformation matrix for a 2-noded traction
!!  element. It determines the coordinates of the element by accessing the
!!  field data for the problem. The matrix has the following form:
!!
!!    T = [  cos(theta)   sin(theta) ]
!!        [ -sin(theta)   cos(theta) ]
!!
!!  where theta = atan( (y2-y1) / (x2-x1) ). The points (x1,y1) and
!!  (x2,y2) are the beginning and end coordinates of the element,
!!  respectively.
! ------------------------------------------------------------------------
SUBROUTINE tshp_transform_2_noded_linear_ (i, T)
  INTEGER, INTENT(IN) :: i
  TYPE(matrixT), INTENT(INOUT) :: T
  DOUBLE PRECISION :: x1,y1, x2,y2    !> element coordinates
  DOUBLE PRECISION :: length          !> length of element
  DOUBLE PRECISION :: c,s             !> for cos(theta) and sin(theta)

  x1 = fld_getCoord( bnd_getConnect(i,1), 1 )
  y1 = fld_getCoord( bnd_getConnect(i,1), 2 )
  x2 = fld_getCoord( bnd_getConnect(i,2), 1 )
  y2 = fld_getCoord( bnd_getConnect(i,2), 2 )

  length = bnd_lenBoundElem(i)

  c = (x2-x1)/length
  s = (y2-y1)/length

  CALL dm_init(T, NDIM,NDIM)

  CALL dm_set(T, 1,1, c)
  CALL dm_set(T, 2,1, -s)
  CALL dm_set(T, 1,2, s)
  CALL dm_set(T, 2,2, c)

END SUBROUTINE tshp_transform_2_noded_linear_



! ************************************************************************
! LOCAL FUNCTIONS
! ************************************************************************

! none

END MODULE traction_element_interpolation
