! ------------------------------------------------------------------------
!>  \brief Module defining solver for linear systems of equations
! ------------------------------------------------------------------------
MODULE linear_solver
USE log_message_control   !> Printing log/error messages
USE log_messages          !> Log/error codes
USE band_sym_matrix_def   !> Banded Symmetric Matrix data type
USE vector_def            !> Vector data type

IMPLICIT NONE
PRIVATE



! ************************************************************************
! EXPORTS
! ************************************************************************

!> Exported interfaces
PUBLIC :: lin_solve



! ************************************************************************
! LOCAL CONSTANTS
! ************************************************************************

!> Sender code for LINear SoLVer module
INTEGER, PARAMETER :: sdr = LINSLV


! ************************************************************************
! DATA TYPES
! ************************************************************************

! none



! ************************************************************************
! INTERFACES
! ************************************************************************


!>  \brief  Interface to linear solver routine
INTERFACE lin_solve
  MODULE PROCEDURE lin_solve_
  MODULE PROCEDURE lin_solve_exc_
END INTERFACE lin_solve



CONTAINS



! ************************************************************************
! ACCESS PROGRAMS
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Linear system of equations solver (non-exception checking)
!!
!!  \param A    Reference a banded symmetric matrix of coefficients
!!  \param b    Reference to a vector of right-hand-sides
!!  \param x    Reference to the vector for the solution (need not be initialized)
!!  \param hbw  Half bandwidth of A (including the diagonal)
!!  \param n    Number of equations (number of rows in A / number of entries in b)
!!  \param info Error code for LAPACK solver (info.EQ.0 -> success | i.LT.0 -> illegal input | i.GT.0 -> A not pos def)
!!
!!  This routine solves the system A*x = b. The matrix A should be
!!  symmetric and positive definite. If the matrix A has not been
!!  previously used in the solution of a system, it computes the Cholesky
!!  decomposition using DPBTRF from the LAPACK library. Once the
!!  decomposition is available it solves the system using DPBTRS, also
!!  from the LAPACK library.
! ------------------------------------------------------------------------
SUBROUTINE lin_solve_ (A,b,x, hbw,n, info)
  TYPE(bandSymMatrixT), INTENT(INOUT) :: A
  TYPE(vectorT), INTENT(IN) :: b
  TYPE(vectorT), INTENT(OUT) :: x
  INTEGER, INTENT(IN) :: hbw,n
  INTEGER, INTENT(OUT) :: info
  CHARACTER, PARAMETER :: uplo = 'u'  !> indicates upper band storage
  INTEGER, PARAMETER :: nrhs = 1      !> number of right-hand-sides
  INTEGER :: kd                       !> number of superdiagonals
  INTEGER :: ldab,ldb                 !> leading dimensions of A and b
  DOUBLE PRECISION, DIMENSION(hbw,n) :: work_mat  !> workspace for matrix
  DOUBLE PRECISION, DIMENSION(n) :: work_vec      !> workspace for vector

  kd = hbw-1    !> number of off-diagonal bands
  ldab = hbw    !> leading dimension of A in banded storage format
  ldb = n       !> leading dimension of right-hand-sides

  !> if A has already been decomposed, just solve the system
  IF (bsm_isDecomposed(A)) THEN

    work_mat = A%decomp   !> set matrix workspace to Cholesky decomposition
    work_vec = b%dat      !> set vector workspace to right-hand-side

  !> if A is not yet decomposed, decompose and then solve
  ELSE

    work_mat = A%dat    !> set matrix workspace to A in banded storage

    !> perform Cholesky decomposition, work_mat is the decomposition on exit
    CALL DPBTRF( uplo, n,kd, work_mat,ldab, info )

    !> if decomposition was successful, set the decomposition in A
    IF (info.EQ.0) THEN
      CALL bsm_setDecomp(A,work_mat)

    !> otherwise, return
    ELSE
      RETURN
    END IF

    work_vec = b%dat  !> set vector workspace to right-hand-side

  END IF


  !> solve the system, work_vec is the solution on exit
  CALL DPBTRS( uplo, n,kd,nrhs, work_mat,ldab, work_vec,ldb, info )

  !> if solution was successful, initialize solution vector and set its data to the solution
  IF (info.EQ.0) THEN
    CALL vec_init(x,n)
    x%dat = work_vec
  END IF

END SUBROUTINE lin_solve_


! ------------------------------------------------------------------------
!>  \brief Linear system of equations solver (exception checking)
!!
!!  \param A    Reference a banded symmetric matrix of coefficients
!!  \param b    Reference to a vector of right-hand-sides
!!  \param x    Reference to the vector for the solution (need not be initialized)
!!  \param hbw  Half bandwidth of A (including the diagonal)
!!  \param n    Number of equations (number of rows in A / number of entries in b)
!!  \param info Error code for LAPACK solver (info.EQ.0 -> success | i.LT.0 -> illegal input | i.GT.0 -> A not pos def)
!!  \param exc  Error code
!!
!!  This routine solves the system A*x = b. The matrix A should be
!!  symmetric and positive definite. If the matrix A has not been
!!  previously used in the solution of a system, it computes the Cholesky
!!  decomposition using DPBTRF from the LAPACK library. Once the
!!  decomposition is available it solves the system using DPBTRS, also
!!  from the LAPACK library. If the number of rows in A does not match
!!  the number of entries in x, it returns a DIMEN exception.
! ------------------------------------------------------------------------
SUBROUTINE lin_solve_exc_ (A,b,x, hbw,n, info, exc)
  TYPE(bandSymMatrixT), INTENT(INOUT) :: A
  TYPE(vectorT), INTENT(IN) :: b
  TYPE(vectorT), INTENT(OUT) :: x
  INTEGER, INTENT(IN) :: hbw,n
  INTEGER, INTENT(OUT) :: info
  INTEGER, INTENT(OUT) :: exc

  !> make sure number of equations in A matches number of right-hand-sides in b
  !> also double-check that input size characteristics match those of A
  IF (bsm_numRows(A).NE.vec_length(b) &
        .OR. bsm_halfBW(A).NE.hbw &
        .OR. bsm_numRows(A).NE.n ) THEN
    exc=DIMEN
    CALL log_printLogMsg(exc,sdr)
    info=-1
    RETURN
  ELSE
    exc=OK
  END IF

  !> call the non-exception checking solver
  CALL lin_solve(A,b,x, hbw,n, info)

  !> if solution was unsuccessful, it means A was not positive definite
  !! (since the selected LAPACK solver only solves systems with banded
  !! symmetric positive definite coefficient matrices, raise POSDEF
  !! exception in this case)
  IF (info.NE.0) THEN
    CALL vec_clean(x)   !> destroy x since it is not meaningful
    exc=POSDEF
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

END SUBROUTINE lin_solve_exc_


END MODULE linear_solver



! ************************************************************************
! LOCAL FUNCTIONS
! ************************************************************************

! none
