! ------------------------------------------------------------------------
!>  \brief Module for testing Linear Solver module
! ------------------------------------------------------------------------
MODULE linear_solver_test
USE fruit                   !> Unit testing framework
USE log_message_control     !> Printing log/error messages
USE log_messages            !> Log/error codes
USE band_sym_matrix_def     !> Banded Symmetric Matrix data type
USE vector_def              !> Vector data type
USE linear_solver           !> Linear Solver module

IMPLICIT NONE


CONTAINS


! ------------------------------------------------------------------------
!>  \test Test for DIMEN exception
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test_A   Test banded symmetric matrix
!!  \param test_b   Test right-hand-side vector
!!  \param test_x   Dummy vector for solution
!!  \param testName Filename for log file (required for exceptions)
!!  \param hbw      Half bandwidth of the matrix
!!  \param n        Number of rows/columns in the matrix
!!  \param expMsg   Expected log message
!!  \param expInfo  Expected LAPACK info code
!!  \param actMsg   Actual log message
!!  \param actInfo  Actual LAPACK info code
!!
!!  This test tries to solve a system of equations when the dimensions of
!!  the coefficient matrix and right-hand-side vector are not compatible.
!!  That is, the number of rows is not the same. It should throw a DIMEN
!!  exception in this case.
! ------------------------------------------------------------------------
SUBROUTINE test_linear_solver_DIMEN
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_linear_solver_DIMEN'
  TYPE(bandSymMatrixT) :: test_A
  TYPE(vectorT) :: test_b, test_x
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: hbw=3,n=10
  INTEGER, PARAMETER :: expMsg=DIMEN, expInfo=-1
  INTEGER :: actMsg, actInfo

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize matrix and vector (data contents not important)
  CALL bsm_init(test_A, hbw,n)
  CALL vec_init(test_b, n+1)    !> note differing dimension

  !> attempt to solve the system
  CALL lin_solve(test_A,test_b,test_x, hbw,n, actInfo, actMsg)
  CALL assertEquals(expInfo,actInfo)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate objects
  CALL log_closeLogFile()
  CALL bsm_clean(test_A)
  CALL vec_clean(test_b)
  CALL vec_clean(test_x)

END SUBROUTINE test_linear_solver_DIMEN


! ------------------------------------------------------------------------
!>  \test Test for POSDEF exception
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test_A   Test banded symmetric matrix (not positive definite)
!!  \param test_b   Test right-hand-side vector
!!  \param test_x   Dummy vector for solution
!!  \param testName Filename for log file (required for exceptions)
!!  \param hbw      Half bandwidth of the matrix
!!  \param n        Number of rows/columns in the matrix
!!  \param expMsg   Expected log message
!!  \param expInfo  Expected LAPACK info code
!!  \param actMsg   Actual log message
!!  \param actInfo  Actual LAPACK info code
!!
!!  This test tries to solve a system of equations when the coefficient
!!  matrix, test_A, is not positive definite. It should throw a POSDEF
!!  exception in this case.
! ------------------------------------------------------------------------
SUBROUTINE test_linear_solver_POSDEF
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_linear_solver_POSDEF'
  TYPE(bandSymMatrixT) :: test_A
  TYPE(vectorT) :: test_b, test_x
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: hbw=3,n=10
  INTEGER, PARAMETER :: expMsg=POSDEF, expInfo=1  !> matrix minor 1 is not pos def
  INTEGER :: actMsg, actInfo
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> build a symmetric, but not positive definite, matrix in test_A
  !! (Matrix is all -2 on the bands, therefore some of the matrix minors
  !! do not have determinant greater than 0. In particular, matrix minor
  !! 1 has negative determinant)
  CALL bsm_init(test_A, hbw,n)
  DO i = 1,n
    DO j = i,MIN(i+hbw-1,n)
      CALL bsm_set(test_A, i,i, -2.d0)
    END DO
  END DO

  !> build test_b = [1..n]
  CALL vec_init(test_b, n)
  DO i = 1,n
    CALL vec_set(test_b, i, DBLE(i))
  END DO

  !> attempt to compute solution
  CALL lin_solve(test_A,test_b, test_x, hbw,n, actInfo, actMsg)
  CALL assertEquals(expInfo,actInfo)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate objects
  CALL log_closeLogFile()
  CALL bsm_clean(test_A)
  CALL vec_clean(test_b)
  CALL vec_clean(test_x)

END SUBROUTINE test_linear_solver_POSDEF


! ------------------------------------------------------------------------
!>  \test Test for correct solution to linear system of equations (general case)
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test_A   Test banded symmetric matrix
!!  \param test_b   Test right-hand-side vector
!!  \param test_x   Vector for solution
!!  \param hbw      Half bandwidth of the matrix
!!  \param n        Number of rows/columns in the matrix
!!  \param expDat   Expected solution
!!  \param toler    Tolerance on solution (floating point error)
!!
!!  This test solves a system of equations, test_A*test_x = test_b, for
!!  test_x in the general case for a banded symmetric positive definite
!!  matrix, test_A. The right-hand-sides, test_b, are computed through
!!  multiplication of test_A and test_x. That is, the input is
!!  test_b = test_A*test_x. Therefore, the expected output from the solver
!!  with inputs test_A and test_b is the original vector test_x.
! ------------------------------------------------------------------------
SUBROUTINE test_linear_solver_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_linear_solver_VAL'
  TYPE(bandSymMatrixT) :: test_A
  TYPE(vectorT) :: test_b, test_x
  INTEGER, PARAMETER :: hbw=3,n=10
  DOUBLE PRECISION, DIMENSION(n) :: expDat
  DOUBLE PRECISION, PARAMETER :: toler = 1.d-14
  INTEGER :: info   !> info code for LAPACK solver
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> build a symmetric positive definite matrix in test_A
  CALL bsm_init(test_A, hbw,n)
  DO i = 1,n
    DO j = i,MIN(i+hbw-1,n)

      IF (i.EQ.j) THEN
        CALL bsm_set(test_A, i,i, ( 2.d0*DBLE(i) ) )  !> i.EQ.j -> 2*i
      ELSE
        CALL bsm_set(test_A, i,j, -1.d0)              !> i.NE.j -> -1
      END IF

    END DO
  END DO

  !> build test_x = [1..n]
  CALL vec_init(test_x, n)
  DO i = 1,n
    CALL vec_set(test_x, i, DBLE(i))
  END DO

  !> compute input RHS from test_A*test_x
  test_b = test_A*test_x

  !> expected solution is test_x
  expDat = test_x%dat

  !> reset test_x
  CALL vec_clean(test_x)

  !> compute actual solution, " test_x = test_A^(-1)*test_b "
  CALL lin_solve(test_A,test_b, test_x, hbw,n, info)

  ! check that solution matches expected result (within tolerance)
  CALL assertEquals(expDat,test_x%dat, n, toler)

  ! clean up memory allocations
  CALL bsm_clean(test_A)
  CALL vec_clean(test_b)
  CALL vec_clean(test_x)

END SUBROUTINE test_linear_solver_VAL


! ------------------------------------------------------------------------
!>  \test Test for correct solution to linear system of equations (identity)
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test_A   Test banded symmetric matrix (identity matrix)
!!  \param test_b   Test right-hand-side vector
!!  \param test_x   Vector for solution
!!  \param hbw      Half bandwidth of the matrix
!!  \param n        Number of rows/columns in the matrix
!!  \param expDat   Expected solution
!!  \param toler    Tolerance on solution (floating point error)
!!
!!  This test solves a system of equations, test_A*test_x = test_b, for
!!  test_x for the case that test_A is the identity matrix. Therefore,
!!  the expected result is that test_x = test_b.
! ------------------------------------------------------------------------
SUBROUTINE test_linear_solver_IDENT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_linear_solver_IDENT'
  TYPE(bandSymMatrixT) :: test_A
  TYPE(vectorT) :: test_b, test_x
  INTEGER, PARAMETER :: hbw=1,n=10
  DOUBLE PRECISION, DIMENSION(n) :: expDat
  DOUBLE PRECISION, PARAMETER :: toler = 1.d-14
  INTEGER :: info   !> info code for LAPACK solver
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> build the identity matrix in test_A
  CALL bsm_init(test_A, hbw,n)
  DO i = 1,n
    CALL bsm_set(test_A, i,i, 1.d0)
  END DO

  !> build test_b = [1..n]
  CALL vec_init(test_b, n)
  DO i = 1,n
    CALL vec_set(test_b, i, DBLE(i))
  END DO

  !> expected solution is test_b
  expDat = test_b%dat

  !> compute actual solution, " test_x = test_A^(-1)*test_b "
  CALL lin_solve(test_A,test_b, test_x, hbw,n, info)

  ! check that solution matches expected result (within tolerance)
  CALL assertEquals(expDat,test_x%dat, n, toler)

  ! clean up memory allocations
  CALL bsm_clean(test_A)
  CALL vec_clean(test_b)
  CALL vec_clean(test_x)

END SUBROUTINE test_linear_solver_IDENT


! ------------------------------------------------------------------------
!>  \test Test for correct solution to linear system of equations (zero)
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test_A   Test banded symmetric matrix
!!  \param test_b   Test right-hand-side vector (zeros)
!!  \param test_x   Vector for solution
!!  \param hbw      Half bandwidth of the matrix
!!  \param n        Number of rows/columns in the matrix
!!  \param expDat   Expected solution
!!  \param toler    Tolerance on solution (floating point error)
!!
!!  This test solves a system of equations, test_A*test_x = test_b, for
!!  test_x for the case that test_b is the zero vector. Therefore,
!!  the expected result is that test_x = test_b = {0}.
! ------------------------------------------------------------------------
SUBROUTINE test_linear_solver_ZERO
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_linear_solver_ZERO'
  TYPE(bandSymMatrixT) :: test_A
  TYPE(vectorT) :: test_b, test_x
  INTEGER, PARAMETER :: hbw=3,n=10
  DOUBLE PRECISION, DIMENSION(n) :: expDat
  DOUBLE PRECISION, PARAMETER :: toler = 1.d-14
  INTEGER :: info   !> info code for LAPACK solver
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> build a symmetric positive definite matrix in test_A
  CALL bsm_init(test_A, hbw,n)
  DO i = 1,n
    DO j = i,MIN(i+hbw-1,n)

      IF (i.EQ.j) THEN
        CALL bsm_set(test_A, i,i, ( 2.d0 * DBLE(i) ) )  !> i.EQ.j -> 2*i
      ELSE
        CALL bsm_set(test_A, i,j, -1.d0)                !> i.NE.j -> -1
      END IF

    END DO
  END DO

  !> initialize test_b (zeros)
  CALL vec_init(test_b, n)

  !> expected solution is test_b
  expDat = test_b%dat

  !> compute actual solution, " test_x = test_A^(-1)*test_b "
  CALL lin_solve(test_A,test_b, test_x, hbw,n, info)

  ! check that solution matches expected result (within tolerance)
  CALL assertEquals(expDat,test_x%dat, n, toler)

  ! clean up memory allocations
  CALL bsm_clean(test_A)
  CALL vec_clean(test_b)
  CALL vec_clean(test_x)

END SUBROUTINE test_linear_solver_ZERO


! ------------------------------------------------------------------------
!>  \test Test for solution in case when decomposition is reused
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test_A   Test banded symmetric matrix
!!  \param test_b   Test right-hand-side vector
!!  \param test_x   Vector for solution
!!  \param hbw      Half bandwidth of the matrix
!!  \param n        Number of rows/columns in the matrix
!!  \param expDat   Expected solution
!!  \param toler    Tolerance on solution (floating point error)
!!
!!  This test solves a system of equations, test_A*test_x = test_b, for
!!  test_x in the general case for a banded symmetric positive definite
!!  matrix, test_A. The right-hand-sides, test_b, are computed through
!!  multiplication of test_A and test_x. That is, the input is
!!  test_b = test_A*test_x. Therefore, the expected output from the solver
!!  with inputs test_A and test_b is the original vector test_x. In
!!  particular, this test solves the system twice. On the second attempt,
!!  the branch of the linear solver that reuses the decomposition of
!!  test_A is activated.
! ------------------------------------------------------------------------
SUBROUTINE test_linear_solver_DECOMP
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_linear_solver_DECOMP'
  TYPE(bandSymMatrixT) :: test_A
  TYPE(vectorT) :: test_b, test_x
  INTEGER, PARAMETER :: hbw=3,n=10
  DOUBLE PRECISION, DIMENSION(n) :: expDat
  DOUBLE PRECISION, PARAMETER :: toler = 1.d-14
  INTEGER :: info   !> info code for LAPACK solver
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> build a symmetric positive definite matrix in test_A
  CALL bsm_init(test_A, hbw,n)
  DO i = 1,n
    DO j = i,MIN(i+hbw-1,n)

      IF (i.EQ.j) THEN
        CALL bsm_set(test_A, i,i, ( 2.d0 * DBLE(i) ) )  !> i.EQ.j -> 2*i
      ELSE
        CALL bsm_set(test_A, i,j, -1.d0)                !> i.NE.j -> -1
      END IF

    END DO
  END DO

  !> build test_x = [1..n]
  CALL vec_init(test_x, n)
  DO i = 1,n
    CALL vec_set(test_x, i, DBLE(i))
  END DO

  !> compute input RHS from test_A*test_x
  test_b = test_A*test_x

  !> expected solution is test_x
  expDat = test_x%dat

  !> reset test_x and solve once, " test_x = test_A^(-1)*test_b "
  CALL vec_clean(test_x)
  CALL lin_solve(test_A,test_b, test_x, hbw,n, info)

  !> reset test_x and solve again
  CALL vec_clean(test_x)
  CALL lin_solve(test_A,test_b, test_x, hbw,n, info)

  ! check that solution matches expected result (within tolerance)
  CALL assertEquals(expDat,test_x%dat, n, toler)

  ! clean up memory allocations
  CALL bsm_clean(test_A)
  CALL vec_clean(test_b)
  CALL vec_clean(test_x)

END SUBROUTINE test_linear_solver_DECOMP

END MODULE linear_solver_test
