! ------------------------------------------------------------------------
!>  \brief Module for Material Property Data
! ------------------------------------------------------------------------
MODULE material_data
USE system_constants        !> Global system constants (for min/max prop values)
USE log_message_control     !> Print log/error messages
USE log_messages            !> Log/error codes and messages

IMPLICIT NONE
PRIVATE



! ************************************************************************
! EXPORTS
! ************************************************************************


!> Exported interfaces
PUBLIC :: mtl_init, mtl_clean, &
          mtl_numMtl, &
          mtl_getEmod, mtl_setEmod, &
          mtl_getPois, mtl_setPois, &
          mtl_getDens, mtl_setDens



! ************************************************************************
! LOCAL CONSTANTS
! ************************************************************************

!> sender code for this module
INTEGER, PARAMETER :: sdr = MTLDAT



! ************************************************************************
! DATA TYPES
! ************************************************************************


!> enumerated type indicating material model
ENUM, BIND(C)
  ENUMERATOR :: linear_elastic
END ENUM


! ------------------------------------------------------------------------
!>  \brief Material Type
!!
!!  \param num    Hash value indicating material number
!!  \param typ    Material model type (from enumerated type in this module)
!!  \param emod   Elastic modulus
!!  \param nu     Poisson's ratio
!!  \param rho    Density
! ------------------------------------------------------------------------
TYPE materialT
  INTEGER :: num
  INTEGER :: typ
  DOUBLE PRECISION :: emod, nu, rho
END TYPE materialT



! ************************************************************************
! STATE VARIABLES
! ************************************************************************

TYPE(materialT), ALLOCATABLE :: materials(:)



! ************************************************************************
! INTERFACES
! ************************************************************************


!> \brief Interface to constructor for module state variables
INTERFACE mtl_init
  MODULE PROCEDURE mtl_init_
  MODULE PROCEDURE mtl_init_exc_
END INTERFACE mtl_init

!> \brief Interface to destructor for module state variables
INTERFACE mtl_clean
  MODULE PROCEDURE mtl_clean_
END INTERFACE mtl_clean

!> \brief Interface for number of materials
INTERFACE mtl_numMtl
  MODULE PROCEDURE mtl_num_mtl_
END INTERFACE mtl_numMtl

!> \brief Interface for getter for elastic modulus
INTERFACE mtl_getEmod
  MODULE PROCEDURE mtl_get_emod_
  MODULE PROCEDURE mtl_get_emod_exc_
END INTERFACE mtl_getEmod

!> \brief Interface for setter for elastic modulus
INTERFACE mtl_setEmod
  MODULE PROCEDURE mtl_set_emod_
  MODULE PROCEDURE mtl_set_emod_exc_
END INTERFACE mtl_setEmod

!> \brief Interface for getter for Poisson's ratio
INTERFACE mtl_getPois
  MODULE PROCEDURE mtl_get_pois_
  MODULE PROCEDURE mtl_get_pois_exc_
END INTERFACE mtl_getPois

!> \brief Interface for setter for Poisson's ratio
INTERFACE mtl_setPois
  MODULE PROCEDURE mtl_set_pois_
  MODULE PROCEDURE mtl_set_pois_exc_
END INTERFACE mtl_setPois

!> \brief Interface for getter for density
INTERFACE mtl_getDens
  MODULE PROCEDURE mtl_get_dens_
  MODULE PROCEDURE mtl_get_dens_exc_
END INTERFACE mtl_getDens

!> \brief Interface for setter for density
INTERFACE mtl_setDens
  MODULE PROCEDURE mtl_set_dens_
  MODULE PROCEDURE mtl_set_dens_exc_
END INTERFACE mtl_setDens



CONTAINS



! ************************************************************************
! ACCESS PROGRAMS
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Constructor for Material Data state variables (non-exception checking)
!!
!!  \param  nmtl  Number of materials
!!
!!  This routine allocates memory and initializes the state variable that
!!  contains the set of material property information for the problem.
! ------------------------------------------------------------------------
SUBROUTINE mtl_init_ (nmtl)
  INTEGER, INTENT(IN) :: nmtl
  INTEGER :: imtl     !> loop variable

  !> only reallocate if new dimension does not match existing dimension
  IF (mtl_numMtl().NE.nmtl) THEN

    !> ensure state variable is clear
    CALL mtl_clean()

    !> allocate memory for material data state variable
    ALLOCATE(materials(nmtl))

  END IF

  !> initialize state variable
  DO imtl = 1,nmtl
    materials(imtl)%num   = imtl
    materials(imtl)%typ   = linear_elastic
    materials(imtl)%emod  = 0.d0
    materials(imtl)%nu    = 0.d0
    materials(imtl)%rho   = 0.d0
  END DO

END SUBROUTINE mtl_init_


! ------------------------------------------------------------------------
!>  \brief Constructor for Material Data state variables (exception checking)
!!
!!  \param      nmtl  Number of materials
!!  \param      exc   Error code
!!
!!  \exception  ALLOC Memory allocation for state variable failed
!!  \exception  SZE   Specified number of materials is invalid
!!
!!  This routine allocates memory and initializes the state variable that
!!  contains the set of material property information for the problem.
! ------------------------------------------------------------------------
SUBROUTINE mtl_init_exc_ (nmtl, exc)
  INTEGER, INTENT(IN) :: nmtl
  INTEGER, INTENT(OUT) :: exc
  INTEGER :: e        !> error code for allocation
  INTEGER :: imtl     !> loop variable

  !> ensure that specified number of materials is valid
  IF (nmtl.LT.1 .OR. nmtl.GT.MAX_MATERIALS) THEN
    exc=SZE
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> only reallocate if new dimension does not match existing dimension
  IF (mtl_numMtl().NE.nmtl) THEN

    !> ensure state variable is clear
    CALL mtl_clean()

    !> allocate memory for material data state variable
    ALLOCATE(materials(nmtl), STAT=e)

    !> ensure that memory allocation was successful
    IF (e.NE.0) THEN
      exc=ALLOC
      CALL log_printLogMsg(exc,sdr)
      RETURN
    ELSE
      exc=OK
    END IF

  END IF

  !> initialize state variable
  DO imtl = 1,nmtl
    materials(imtl)%num   = imtl
    materials(imtl)%typ   = linear_elastic
    materials(imtl)%emod  = 0.d0
    materials(imtl)%nu    = 0.d0
    materials(imtl)%rho   = 0.d0
  END DO

END SUBROUTINE mtl_init_exc_


! ------------------------------------------------------------------------
!>  \brief Destructor for Material Data state variables
!!
!!  This routine clears the memory allocated to the state variable that
!!  contains the set of material property information for the problem.
! ------------------------------------------------------------------------
SUBROUTINE mtl_clean_ ()

  IF (ALLOCATED(materials)) DEALLOCATE(materials)

END SUBROUTINE mtl_clean_


! ------------------------------------------------------------------------
!>  \brief Getter for number of materials
!!
!!  \return nmtl  Number of materials
!!
!!  This routine determines the number of data entries that have been
!!  allocated for material data. It does not check that the material data
!!  has been populated (i.e. changed from initial zero values).
! ------------------------------------------------------------------------
FUNCTION mtl_num_mtl_ () RESULT(nmtl)
  INTEGER :: nmtl

  !> if data is initialized, return number of materials
  IF (ALLOCATED(materials)) THEN
    nmtl = SIZE(materials)
  ELSE
    nmtl = 0   !> if not initialized, there are no materials
  END IF

END FUNCTION mtl_num_mtl_


! ------------------------------------------------------------------------
!>  \brief Getter for elastic modulus (non-exception checking)
!!
!!  \param  i     Material number
!!  \return emod  Elastic modulus
!!
!!  This routine determines the value of the elastic modulus for material
!!  number i.
! ------------------------------------------------------------------------
FUNCTION mtl_get_emod_ (i) RESULT(emod)
  INTEGER, INTENT(IN) :: i
  DOUBLE PRECISION :: emod

  emod = materials(i)%emod

END FUNCTION mtl_get_emod_


! ------------------------------------------------------------------------
!>  \brief Getter for elastic modulus (exception checking)
!!
!!  \param  i     Material number
!!  \param  exc   Error code
!!
!!  \return emod  Elastic modulus
!!
!!  \exception  POSIT The material number is not in [1..mtl_numMtls()]
!!  \exception  TYP   The material type is not 'linear_elastic'
!!
!!  This routine determines the value of the elastic modulus for material
!!  number i.
! ------------------------------------------------------------------------
FUNCTION mtl_get_emod_exc_ (i, exc) RESULT(emod)
  INTEGER, INTENT(IN) :: i
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: emod

  !> check that the index is within the bounds of the material list
  IF ( i.LT.1 .OR. i.GT.mtl_numMtl() ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    emod = 0.d0
    RETURN

  !> check that the material type is linear elastic
  !! (otherwise, elastic modulus may not be appropriate)
  ELSE IF ( materials(i)%typ .NE. linear_elastic ) THEN
    exc=TYP
    CALL log_printLogMsg(exc,sdr)
    emod = 0.d0
    RETURN

  ELSE
    exc=OK
  END IF

  !> call non-exception version
  emod = mtl_getEmod(i)

END FUNCTION mtl_get_emod_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for elastic modulus (non-exception checking)
!!
!!  \param  i     Material number
!!  \param  emod  Elastic modulus
!!
!!  This routine sets the value of the elastic modulus for material
!!  number i.
! ------------------------------------------------------------------------
SUBROUTINE mtl_set_emod_ (i,emod)
  INTEGER, INTENT(IN) :: i
  DOUBLE PRECISION, INTENT(IN) :: emod

  materials(i)%emod = emod

END SUBROUTINE mtl_set_emod_


! ------------------------------------------------------------------------
!>  \brief Setter for elastic modulus (exception checking)
!!
!!  \param  i     Material number
!!  \param  emod  Elastic modulus
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  The value of elastic modulus exceeds defined limits
!!  \exception  POSIT   The material number is not in [1..mtl_numMtls()]
!!  \exception  TYP     The material type is not 'linear_elastic'
!!
!!  This routine sets the value of the elastic modulus for material
!!  number i.
! ------------------------------------------------------------------------
SUBROUTINE mtl_set_emod_exc_ (i,emod, exc)
  INTEGER, INTENT(IN) :: i
  DOUBLE PRECISION, INTENT(IN) :: emod
  INTEGER, INTENT(OUT) :: exc

  !> check that the index is within the bounds of the material list
  IF ( i.LT.1 .OR. i.GT.mtl_numMtl() ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  !> check that the material type is linear elastic
  !! (otherwise, elastic modulus may not be appropriate)
  ELSE IF ( materials(i)%typ .NE. linear_elastic ) THEN
    exc=TYP
    CALL log_printLogMsg(exc,sdr)
    RETURN

  !> check that the value of elastic modulus is within prescribed limits
  ELSE IF (emod.LT.E_MIN .OR. emod.GT.E_MAX) THEN
    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK

  END IF

  !> call non-exception version
  CALL mtl_setEmod(i,emod)

END SUBROUTINE mtl_set_emod_exc_


! ------------------------------------------------------------------------
!>  \brief Getter for Poisson's ratio (non-exception checking)
!!
!!  \param  i   Material number
!!  \return nu  Poisson's ratio
!!
!!  This routine determines the value of Poisson's ratio for material
!!  number i.
! ------------------------------------------------------------------------
FUNCTION mtl_get_pois_ (i) RESULT(nu)
  INTEGER, INTENT(IN) :: i
  DOUBLE PRECISION :: nu

  nu = materials(i)%nu

END FUNCTION mtl_get_pois_


! ------------------------------------------------------------------------
!>  \brief Getter for Poisson's ratio (exception checking)
!!
!!  \param  i     Material number
!!  \param  exc   Error code
!!
!!  \return nu    Poisson's ratio
!!
!!  \exception  POSIT The material number is not in [1..mtl_numMtls()]
!!  \exception  TYP   The material type is not 'linear_elastic'
!!
!!  This routine determines the value of Poisson's ratio for material
!!  number i.
! ------------------------------------------------------------------------
FUNCTION mtl_get_pois_exc_ (i, exc) RESULT(nu)
  INTEGER, INTENT(IN) :: i
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: nu

  !> check that the index is within the bounds of the material list
  IF ( i.LT.1 .OR. i.GT.mtl_numMtl() ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    nu = 0.d0
    RETURN

  !> check that the material type is linear elastic
  !! (otherwise, Poisson's ratio may not be appropriate)
  ELSE IF ( materials(i)%typ .NE. linear_elastic ) THEN
    exc=TYP
    CALL log_printLogMsg(exc,sdr)
    nu = 0.d0
    RETURN

  ELSE
    exc=OK
  END IF

  !> call non-exception version
  nu = mtl_getPois(i)

END FUNCTION mtl_get_pois_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for Poisson's ratio (non-exception checking)
!!
!!  \param  i     Material number
!!  \param  nu    Poisson's ratio
!!
!!  This routine sets the value of Poisson's ratio for material
!!  number i.
! ------------------------------------------------------------------------
SUBROUTINE mtl_set_pois_ (i,nu)
  INTEGER, INTENT(IN) :: i
  DOUBLE PRECISION, INTENT(IN) :: nu

  materials(i)%nu = nu

END SUBROUTINE mtl_set_pois_


! ------------------------------------------------------------------------
!>  \brief Setter for Poisson's ratio (exception checking)
!!
!!  \param  i     Material number
!!  \param  nu    Poisson's ratio
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  The value of Poisson's ratio exceeds defined limits
!!  \exception  POSIT   The material number is not in [1..mtl_numMtls()]
!!  \exception  TYP     The material type is not 'linear_elastic'
!!
!!  This routine sets the value of Poisson's ratio for material
!!  number i.
! ------------------------------------------------------------------------
SUBROUTINE mtl_set_pois_exc_ (i,nu, exc)
  INTEGER, INTENT(IN) :: i
  DOUBLE PRECISION, INTENT(IN) :: nu
  INTEGER, INTENT(OUT) :: exc

  !> check that the index is within the bounds of the material list
  IF ( i.LT.1 .OR. i.GT.mtl_numMtl() ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  !> check that the material type is linear elastic
  !! (otherwise, elastic modulus may not be appropriate)
  ELSE IF ( materials(i)%typ .NE. linear_elastic ) THEN
    exc=TYP
    CALL log_printLogMsg(exc,sdr)
    RETURN

  !> check that the value of elastic modulus is within prescribed limits
  ELSE IF (nu.LT.NU_MIN .OR. nu.GT.NU_MAX) THEN
    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK

  END IF

  !> call non-exception version
  CALL mtl_setPois(i,nu)

END SUBROUTINE mtl_set_pois_exc_



! ------------------------------------------------------------------------
!>  \brief Getter for density (non-exception checking)
!!
!!  \param  i     Material number
!!  \return rho   Density
!!
!!  This routine determines the value of the density of material
!!  number i.
! ------------------------------------------------------------------------
FUNCTION mtl_get_dens_ (i) RESULT(rho)
  INTEGER, INTENT(IN) :: i
  DOUBLE PRECISION :: rho

  rho = materials(i)%rho

END FUNCTION mtl_get_dens_


! ------------------------------------------------------------------------
!>  \brief Getter for density (exception checking)
!!
!!  \param  i     Material number
!!  \param  exc   Error code
!!
!!  \return rho    Density
!!
!!  \exception  POSIT The material number is not in [1..mtl_numMtls()]
!!
!!  This routine determines the value of the density of material
!!  number i.
! ------------------------------------------------------------------------
FUNCTION mtl_get_dens_exc_ (i, exc) RESULT(rho)
  INTEGER, INTENT(IN) :: i
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: rho

  !> check that the index is within the bounds of the material list
  IF ( i.LT.1 .OR. i.GT.mtl_numMtl() ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    rho = 0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  rho = mtl_getDens(i)

END FUNCTION mtl_get_dens_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for density (non-exception checking)
!!
!!  \param  i     Material number
!!  \param  rho   Density
!!
!!  This routine sets the value of the density of material number i.
! ------------------------------------------------------------------------
SUBROUTINE mtl_set_dens_ (i,rho)
  INTEGER, INTENT(IN) :: i
  DOUBLE PRECISION, INTENT(IN) :: rho

  materials(i)%rho = rho

END SUBROUTINE mtl_set_dens_


! ------------------------------------------------------------------------
!>  \brief Setter for density (exception checking)
!!
!!  \param  i     Material number
!!  \param  rho   Density
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  The value of the density exceeds defined limits
!!  \exception  POSIT   The material number is not in [1..mtl_numMtls()]
!!
!!  This routine sets the value of the density of material number i.
! ------------------------------------------------------------------------
SUBROUTINE mtl_set_dens_exc_ (i,rho, exc)
  INTEGER, INTENT(IN) :: i
  DOUBLE PRECISION, INTENT(IN) :: rho
  INTEGER, INTENT(OUT) :: exc

  !> check that the index is within the bounds of the material list
  IF ( i.LT.1 .OR. i.GT.mtl_numMtl() ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  !> check that the value of density is within prescribed limits
  ELSE IF (rho.LT.RHO_MIN .OR. rho.GT.RHO_MAX) THEN
    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK

  END IF

  !> call non-exception version
  CALL mtl_setDens(i,rho)

END SUBROUTINE mtl_set_dens_exc_



! ************************************************************************
! LOCAL FUNCTIONS
! ************************************************************************

! none

END MODULE material_data
