! ------------------------------------------------------------------------
!>  \brief Module for testing Material Property Data module
! ------------------------------------------------------------------------
MODULE material_data_test
USE fruit                   !> Unit testing framework
USE system_constants        !> Global constants
USE log_message_control     !> Printing log/error messages
USE log_messages            !> Log/error codes
USE material_data           !> Material Property Data module

IMPLICIT NONE


CONTAINS


! ------------------------------------------------------------------------
!>  \test Test for OK exception message on allocation
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nmtl       Number of materials
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!
!!  This test initializes the material_data module and makes sure that the
!!  exception message is OK (i.e. allocation did not fail).
! ------------------------------------------------------------------------
SUBROUTINE test_mtl_allocation_MSG
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_mtl_allocation_MSG'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nmtl=10
  INTEGER, PARAMETER :: expMsg=OK
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log message file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize module and check the exception
  CALL mtl_init(nmtl, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate module
  CALL log_closeLogFile()
  CALL mtl_clean()

END SUBROUTINE test_mtl_allocation_MSG


! ------------------------------------------------------------------------
!>  \test Test for SZE exception message on allocation
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!
!!  This test attempts to initialize the material data module with invalid
!!  size parameters and verifies that the correct exception is returned.
! ------------------------------------------------------------------------
SUBROUTINE test_mtl_allocation_SZE
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_mtl_allocation_SZE'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: expMsg=SZE
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> try to initialize with nmtl=0
  CALL mtl_init(0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> ensure module is reset
  CALL mtl_clean()

  !> try to initialize with nmtl=MAX_MATERIALS+1
  CALL mtl_init(MAX_MATERIALS+1, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate module
  CALL log_closeLogFile()
  CALL mtl_clean()

END SUBROUTINE test_mtl_allocation_SZE


! ------------------------------------------------------------------------
!>  \test Test for number of materials when module is not initialized
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param expected     Expected number of materials
!!  \param actual       Actual number of materials
!!
!!  This test makes sure that the number of materials is returned as 0
!!  when the module is not initialized
! ------------------------------------------------------------------------
SUBROUTINE test_mtl_num_mtl_not_allocated
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_mtl_num_mtl_not_allocated'
  INTEGER, PARAMETER :: expected = 0
  INTEGER :: actual

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> check number of materials
  actual = mtl_numMtl()
  CALL assertEquals(expected, actual)

END SUBROUTINE test_mtl_num_mtl_not_allocated


! ------------------------------------------------------------------------
!>  \test Test for number of materials when module is initialized
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param expected     Expected number of materials
!!  \param actual       Actual number of materials
!!
!!  This test makes sure that the correct number of materials is returned.
! ------------------------------------------------------------------------
SUBROUTINE test_mtl_num_mtl_allocated
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_mtl_num_mtl_allocated'
  INTEGER, PARAMETER :: expected = 10
  INTEGER :: actual

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize the module
  CALL mtl_init(expected)

  !> check number of materials
  actual = mtl_numMtl()
  CALL assertEquals(expected, actual)

  !> deallocate the module
  CALL mtl_clean()

END SUBROUTINE test_mtl_num_mtl_allocated


! ------------------------------------------------------------------------
!>  \test Test for POSIT exception from mtl_getEmod
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nmtl       Number of materials
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!  \param emod       Dummy variable for get function return
!!
!!  This test checks that a POSIT exception is returned when the requested
!!  location is not inside the material list.
! ------------------------------------------------------------------------
SUBROUTINE test_mtl_get_emod_POSIT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_mtl_get_emod_POSIT'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nmtl=10
  INTEGER, PARAMETER :: expMsg = POSIT
  INTEGER :: actMsg
  DOUBLE PRECISION :: emod

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize materials
  CALL mtl_init(nmtl)

  !> try to get beyond last material
  emod = mtl_getEmod(nmtl+1, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to get before first material
  emod = mtl_getEmod(0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate the material list
  CALL log_closeLogFile()
  CALL mtl_clean()

END SUBROUTINE test_mtl_get_emod_POSIT


! ------------------------------------------------------------------------
!>  \test Test for POSIT exception from mtl_setEmod
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nmtl       Number of materials
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!  \param emod       Dummy variable for set routine input
!!
!!  This test checks that a POSIT exception is returned when the requested
!!  location is not inside the material list.
! ------------------------------------------------------------------------
SUBROUTINE test_mtl_set_emod_POSIT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_mtl_set_emod_POSIT'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nmtl=10
  INTEGER, PARAMETER :: expMsg = POSIT
  INTEGER :: actMsg
  DOUBLE PRECISION, PARAMETER :: emod = 3.d0

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize materials
  CALL mtl_init(nmtl)

  !> try to set beyond last material
  CALL mtl_setEmod(nmtl+1,emod, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set before first material
  CALL mtl_setEmod(0,emod, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate the material list
  CALL log_closeLogFile()
  CALL mtl_clean()

END SUBROUTINE test_mtl_set_emod_POSIT


! ------------------------------------------------------------------------
!>  \test Test for EXCEED exception from mtl_setEmod
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nmtl       Number of materials
!!  \param i          Material number for test location
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!
!!  This test checks that an EXCEED exception is returned when the
!!  specified input is not within the range defined in the System
!!  Constants module.
! ------------------------------------------------------------------------
SUBROUTINE test_mtl_set_emod_EXCEED
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_mtl_set_emod_EXCEED'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nmtl=10
  INTEGER, PARAMETER :: i=3
  INTEGER, PARAMETER :: expMsg = EXCEED
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize materials
  CALL mtl_init(nmtl)

  !> try to set below min value
  CALL mtl_setEmod(i,E_MIN-1.d0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set above max value
  CALL mtl_setEmod(i,E_MAX+1.d0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate the material list
  CALL log_closeLogFile()
  CALL mtl_clean()

END SUBROUTINE test_mtl_set_emod_EXCEED


! ------------------------------------------------------------------------
!>  \test Test for correct value getting and setting in mtl_getEmod and mtl_setEmod
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param nmtl       Number of materials
!!  \param i          Material number for test location
!!  \param expVal     Expected data value
!!  \param actVal     Actual data value
!!
!!  This test checks that the correct value is set using the mtl_setEmod
!!  access program and returned from the mtl_getEmod access program.
! ------------------------------------------------------------------------
SUBROUTINE test_mtl_get_set_emod_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_mtl_get_set_emod_VAL'
  INTEGER, PARAMETER :: nmtl=10
  INTEGER, PARAMETER :: i=3
  DOUBLE PRECISION, PARAMETER :: expVal = 3.d0
  DOUBLE PRECISION :: actVal

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize materials
  CALL mtl_init(nmtl)

  !> set the value of the material
  CALL mtl_setEmod(i,expVal)

  !> get the value using the access program
  actVal = mtl_getEmod(i)
  CALL assertEquals(expVal,actVal)

  !> deallocate the materials
  CALL mtl_clean()

END SUBROUTINE test_mtl_get_set_emod_VAL


! ------------------------------------------------------------------------
!>  \test Test for POSIT exception from mtl_getPois
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nmtl       Number of materials
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!  \param nu         Dummy variable for get function return
!!
!!  This test checks that a POSIT exception is returned when the requested
!!  location is not inside the material list.
! ------------------------------------------------------------------------
SUBROUTINE test_mtl_get_pois_POSIT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_mtl_get_pois_POSIT'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nmtl=10
  INTEGER, PARAMETER :: expMsg = POSIT
  INTEGER :: actMsg
  DOUBLE PRECISION :: nu

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize materials
  CALL mtl_init(nmtl)

  !> try to get beyond last material
  nu = mtl_getPois(nmtl+1, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to get before first material
  nu = mtl_getPois(0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate the material list
  CALL log_closeLogFile()
  CALL mtl_clean()

END SUBROUTINE test_mtl_get_pois_POSIT


! ------------------------------------------------------------------------
!>  \test Test for POSIT exception from mtl_setPois
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nmtl       Number of materials
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!  \param nu         Dummy variable for set routine input
!!
!!  This test checks that a POSIT exception is returned when the requested
!!  location is not inside the material list.
! ------------------------------------------------------------------------
SUBROUTINE test_mtl_set_pois_POSIT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_mtl_set_pois_POSIT'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nmtl=10
  INTEGER, PARAMETER :: expMsg = POSIT
  INTEGER :: actMsg
  DOUBLE PRECISION, PARAMETER :: nu = 0.25d0

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize materials
  CALL mtl_init(nmtl)

  !> try to set beyond last material
  CALL mtl_setPois(nmtl+1,nu, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set before first material
  CALL mtl_setPois(0,nu, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate the material list
  CALL log_closeLogFile()
  CALL mtl_clean()

END SUBROUTINE test_mtl_set_pois_POSIT


! ------------------------------------------------------------------------
!>  \test Test for EXCEED exception from mtl_setPois
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nmtl       Number of materials
!!  \param i          Material number for test location
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!
!!  This test checks that an EXCEED exception is returned when the
!!  specified input is not within the range defined in the System
!!  Constants module.
! ------------------------------------------------------------------------
SUBROUTINE test_mtl_set_pois_EXCEED
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_mtl_set_pois_EXCEED'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nmtl=10
  INTEGER, PARAMETER :: i=3
  INTEGER, PARAMETER :: expMsg = EXCEED
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize materials
  CALL mtl_init(nmtl)

  !> try to set below min value
  CALL mtl_setPois(i,NU_MIN-0.1d0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set above max value
  CALL mtl_setPois(i,NU_MAX+0.1d0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate the material list
  CALL log_closeLogFile()
  CALL mtl_clean()

END SUBROUTINE test_mtl_set_pois_EXCEED


! ------------------------------------------------------------------------
!>  \test Test for correct value getting and setting in mtl_getPois and mtl_setPois
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param nmtl       Number of materials
!!  \param i          Material number for test location
!!  \param expVal     Expected data value
!!  \param actVal     Actual data value
!!
!!  This test checks that the correct value is set using the mtl_setPois
!!  access program and returned from the mtl_getPois access program.
! ------------------------------------------------------------------------
SUBROUTINE test_mtl_get_set_pois_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_mtl_get_set_pois_VAL'
  INTEGER, PARAMETER :: nmtl=10
  INTEGER, PARAMETER :: i=3
  DOUBLE PRECISION, PARAMETER :: expVal = 0.25d0
  DOUBLE PRECISION :: actVal

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize materials
  CALL mtl_init(nmtl)

  !> set the value of the material
  CALL mtl_setPois(i,expVal)

  !> get the value using the access program
  actVal = mtl_getPois(i)
  CALL assertEquals(expVal,actVal)

  !> deallocate the materials
  CALL mtl_clean()

END SUBROUTINE test_mtl_get_set_pois_VAL


! ------------------------------------------------------------------------
!>  \test Test for POSIT exception from mtl_getDens
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nmtl       Number of materials
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!  \param rho        Dummy variable for get function return
!!
!!  This test checks that a POSIT exception is returned when the requested
!!  location is not inside the material list.
! ------------------------------------------------------------------------
SUBROUTINE test_mtl_get_dens_POSIT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_mtl_get_dens_POSIT'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nmtl=10
  INTEGER, PARAMETER :: expMsg = POSIT
  INTEGER :: actMsg
  DOUBLE PRECISION :: rho

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize materials
  CALL mtl_init(nmtl)

  !> try to get beyond last material
  rho = mtl_getDens(nmtl+1, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to get before first material
  rho = mtl_getDens(0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate the material list
  CALL log_closeLogFile()
  CALL mtl_clean()

END SUBROUTINE test_mtl_get_dens_POSIT


! ------------------------------------------------------------------------
!>  \test Test for POSIT exception from mtl_setDens
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nmtl       Number of materials
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!  \param rho        Dummy variable for set routine input
!!
!!  This test checks that a POSIT exception is returned when the requested
!!  location is not inside the material list.
! ------------------------------------------------------------------------
SUBROUTINE test_mtl_set_dens_POSIT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_mtl_set_dens_POSIT'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nmtl=10
  INTEGER, PARAMETER :: expMsg = POSIT
  INTEGER :: actMsg
  DOUBLE PRECISION, PARAMETER :: rho = 2.d3

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize materials
  CALL mtl_init(nmtl)

  !> try to set beyond last material
  CALL mtl_setDens(nmtl+1,rho, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set before first material
  CALL mtl_setDens(0,rho, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate the material list
  CALL log_closeLogFile()
  CALL mtl_clean()

END SUBROUTINE test_mtl_set_dens_POSIT


! ------------------------------------------------------------------------
!>  \test Test for EXCEED exception from mtl_setDens
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param testName   Filename for log file (required for exceptions)
!!  \param nmtl       Number of materials
!!  \param i          Material number for test location
!!  \param expMsg     Expected log message
!!  \param actMsg     Actual log message
!!
!!  This test checks that an EXCEED exception is returned when the
!!  specified input is not within the range defined in the System
!!  Constants module.
! ------------------------------------------------------------------------
SUBROUTINE test_mtl_set_dens_EXCEED
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_mtl_set_dens_EXCEED'
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: nmtl=10
  INTEGER, PARAMETER :: i=3
  INTEGER, PARAMETER :: expMsg = EXCEED
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize materials
  CALL mtl_init(nmtl)

  !> try to set below min value
  CALL mtl_setDens(i,RHO_MIN-1.d0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set above max value
  CALL mtl_setDens(i,RHO_MAX+1.d0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate the material list
  CALL log_closeLogFile()
  CALL mtl_clean()

END SUBROUTINE test_mtl_set_dens_EXCEED


! ------------------------------------------------------------------------
!>  \test Test for correct value getting and setting in mtl_getDens and mtl_setDens
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param nmtl       Number of materials
!!  \param i          Material number for test location
!!  \param expVal     Expected data value
!!  \param actVal     Actual data value
!!
!!  This test checks that the correct value is set using the mtl_setDens
!!  access program and returned from the mtl_getDens access program.
! ------------------------------------------------------------------------
SUBROUTINE test_mtl_get_set_dens_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_mtl_get_set_dens_VAL'
  INTEGER, PARAMETER :: nmtl=10
  INTEGER, PARAMETER :: i=3
  DOUBLE PRECISION, PARAMETER :: expVal = 2.d3
  DOUBLE PRECISION :: actVal

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize materials
  CALL mtl_init(nmtl)

  !> set the value of the material
  CALL mtl_setDens(i,expVal)

  !> get the value using the access program
  actVal = mtl_getDens(i)
  CALL assertEquals(expVal,actVal)

  !> deallocate material data
  CALL mtl_clean()

END SUBROUTINE test_mtl_get_set_dens_VAL


END MODULE material_data_test
