! ------------------------------------------------------------------------
!>  \brief Module for testing Vector data type
! ------------------------------------------------------------------------
MODULE vector_test
USE fruit                   !> Unit testing framework
USE log_message_control     !> Printing log/error messages
USE log_messages            !> Log/error codes
USE vector_def              !> Vector data type

IMPLICIT NONE



CONTAINS



! ------------------------------------------------------------------------
!>  \test Test for OK exception message on allocation
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test     Test vector
!!  \param testName Filename for log file (required for exceptions)
!!  \param n        Number of entries in the vector
!!  \param expMsg   Expected log message
!!  \param actMsg   Actual log message
!!
!!  This test initializes a vectorT and makes sure that the exception
!!  message is OK (i.e. allocation did not fail).
! ------------------------------------------------------------------------
SUBROUTINE test_vec_allocation_MSG
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_allocation_MSG'
  TYPE(vectorT) :: test
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: n=3
  INTEGER, PARAMETER :: expMsg=OK
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log message file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize vector and check the exception
  CALL vec_init(test, n, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate vector
  CALL log_closeLogFile()
  CALL vec_clean(test)

END SUBROUTINE test_vec_allocation_MSG


! ------------------------------------------------------------------------
!>  \test Test for SZE exception message on allocation
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test     Test vector
!!  \param testName Filename for log file (required for exceptions)
!!  \param n        Number of entries in the vector
!!  \param expMsg   Expected log message
!!  \param actMsg   Actual log message
!!
!!  This test attempts to initialize a vectorT with invalid size
!!  parameters and verifies that the correct exception is returned.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_allocation_SZE
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_allocation_SZE'
  TYPE(vectorT) :: test
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: n=-1
  INTEGER, PARAMETER :: expMsg=SZE
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log message file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> try to initialize with n=-1
  CALL vec_init(test, n, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate vector
  CALL log_closeLogFile()
  CALL vec_clean(test)

END SUBROUTINE test_vec_allocation_SZE


! ------------------------------------------------------------------------
!>  \test Test for initialization of vector data to zeros.
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test     Test dense matrix
!!  \param n        Number of entries in the vector
!!  \param expDat   Expected initial contents of vector data
!!
!!  This test initializes a vectorT and ensures that the data is
!!  initialized to zeros.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_allocation_DAT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_allocation_DAT'
  TYPE(vectorT) :: test
  INTEGER, PARAMETER :: n=3
  DOUBLE PRECISION, DIMENSION(n) :: expDat

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set up the expected data
  expDat = RESHAPE(  (/ 0.d0,0.d0,0.d0 /), SHAPE(expDat) )

  !> initialize vector and test actual data
  CALL vec_init(test, n)
  CALL assertEquals(expDat,test%dat, n)

  !> deallocate vector
  CALL vec_clean(test)

END SUBROUTINE test_vec_allocation_DAT


! ------------------------------------------------------------------------
!>  \test Test for deallocation
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test       Test vector
!!  \param n          Number of entries in the vector
!!  \param expBefore  Expected allocation status before deallocation
!!  \param expAfter   Expected allocation status after deallocation
!!  \param actBefore  Actual allocation status before deallocation
!!  \param actAfter   Actual allocation status after deallocation
!!
!!  This test initializes a vectorT and ensures that it is allocated. It
!!  then deallocates the object and ensures that it has been deallocated.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_deallocation
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_deallocation'
  TYPE(vectorT) :: test
  INTEGER, PARAMETER :: n=3
  LOGICAL, PARAMETER :: expBefore = .TRUE.
  LOGICAL, PARAMETER :: expAfter = .FALSE.
  LOGICAL :: actBefore, actAfter

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize the vector
  CALL vec_init(test, n)

  !> make sure dat is allocated
  actBefore = ALLOCATED(test%dat)
  CALL assertEquals(expBefore,actBefore)

  !> deallocate the vector
  CALL vec_clean(test)

  !> make sure dat is deallocated
  actAfter = ALLOCATED(test%dat)
  CALL assertEquals(expAfter,actAfter)

END SUBROUTINE test_vec_deallocation


! ------------------------------------------------------------------------
!>  \test Test for length when vector is not allocated
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test       Test vector
!!  \param expected   Expected length
!!  \param actual     Actual length
!!
!!  This test makes sure that the length is returned as 0 when
!!  the vector is not allocated
! ------------------------------------------------------------------------
SUBROUTINE test_vec_length_not_allocated
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_length_not_allocated'
  TYPE(vectorT) :: test
  INTEGER, PARAMETER :: expected = 0
  INTEGER :: actual

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> check length
  actual = vec_length(test)
  CALL assertEquals(expected, actual)

END SUBROUTINE test_vec_length_not_allocated


! ------------------------------------------------------------------------
!>  \test Test for length when vector is allocated
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test       Test vector
!!  \param expected   Expected length
!!  \param actual     Actual length
!!
!!  This test checks that the length returned is correct when the
!!  vector is allocated.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_length_allocated
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_length_allocated'
  TYPE(vectorT) :: test
  INTEGER, PARAMETER :: n=3
  INTEGER, PARAMETER :: expected = n
  INTEGER :: actual

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize the vector
  CALL vec_init(test, n)

  !> check length
  actual = vec_length(test)
  CALL assertEquals(expected, actual)

  !> deallocate the matrix
  CALL vec_clean(test)

END SUBROUTINE test_vec_length_allocated


! ------------------------------------------------------------------------
!>  \test Test for POSIT exception from vec_get
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test     Test vector
!!  \param testName Filename for log file (required for exceptions)
!!  \param m        Number of entries in the vector
!!  \param expMsg   Expected log message
!!  \param actMsg   Actual log message
!!  \param v        Dummy variable for get function return
!!
!!  This test checks that a POSIT exception is returned when the requested
!!  location is not inside the vector.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_get_POSIT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_get_POSIT'
  TYPE(vectorT) :: test
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: m=3
  INTEGER, PARAMETER :: expMsg = POSIT
  INTEGER :: actMsg
  DOUBLE PRECISION :: v

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize vector
  CALL vec_init(test, m)

  !> try to get beyond last entry
  v = vec_get(test, m+1, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to get before first entry
  v = vec_get(test, 0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate the vector
  CALL log_closeLogFile()
  CALL vec_clean(test)

END SUBROUTINE test_vec_get_POSIT


! ------------------------------------------------------------------------
!>  \test Test for correct value return from vec_get
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test     Test vector
!!  \param m        Number of entries in the vector
!!  \param i        Index for test location
!!  \param expVal   Expected data value
!!  \param actVal   Actual data value
!!
!!  This test checks that the correct value is returned from the get
!!  access program.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_get_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_get_VAL'
  TYPE(vectorT) :: test
  INTEGER, PARAMETER :: m=3
  INTEGER, PARAMETER :: i=2
  DOUBLE PRECISION, PARAMETER :: expVal = 3.d0
  DOUBLE PRECISION :: actVal

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize the vector
  CALL vec_init(test, m)

  !> set the value manually in the internal data structure
  test%dat(i) = expVal

  !> get the value using the access program
  actVal = vec_get(test, i)
  CALL assertEquals(expVal,actVal)

  !> deallocate the vector
  CALL vec_clean(test)

END SUBROUTINE test_vec_get_VAL


! ------------------------------------------------------------------------
!>  \test Test for POSIT exception from vec_set
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test     Test vector
!!  \param testName Filename for log file (required for exceptions)
!!  \param m        Number of entries in the vector
!!  \param expMsg   Expected log message
!!  \param actMsg   Actual log message
!!  \param v        Dummy variable for input to set routine
!!
!!  This test checks that a POSIT exception is returned when the requested
!!  location is not inside the vector.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_set_POSIT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_set_POSIT'
  TYPE(vectorT) :: test
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: m=3
  INTEGER, PARAMETER :: expMsg = POSIT
  DOUBLE PRECISION, PARAMETER :: v=0.d0
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize the log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize the vector
  CALL vec_init(test, m)

  !> try to set after last entry
  CALL vec_set(test, m+1,v, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set before first entry
  CALL vec_set(test, 0,v, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate vector
  CALL log_closeLogFile()
  CALL vec_clean(test)

END SUBROUTINE test_vec_set_POSIT


! ------------------------------------------------------------------------
!>  \test Test for correct value return after vec_set
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test     Test vector
!!  \param m        Number of entries in the vector
!!  \param i        Index for test location
!!  \param expVal   Expected data value
!!  \param actVal   Actual data value
!!
!!  This test checks that the correct value is entered by the set
!!  access program.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_set_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_set_VAL'
  TYPE(vectorT) :: test
  INTEGER, PARAMETER :: m=3
  INTEGER, PARAMETER :: i=2
  DOUBLE PRECISION, PARAMETER :: expVal = 3.d0
  DOUBLE PRECISION :: actVal

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize the vector
  CALL vec_init(test, m)

  !> set the value
  CALL vec_set(test, i,expVal)
  actVal = vec_get(test, i)
  CALL assertEquals(expVal,actVal)

  !> deallocate vector
  CALL vec_clean(test)

END SUBROUTINE test_vec_set_VAL


! ------------------------------------------------------------------------
!>  \test Test for DIMEN exception when adding vectors
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      First input vector
!!  \param test2      Second input vector
!!  \param test3      Output vector
!!  \param testName   Filename for log file (required for exceptions)
!!  \param m          Number of entries in the vector
!!  \param expMsg     Expected exception message
!!  \param actMsg     Actual exception message
!!
!!  This test checks that a DIMEN exception is raised when vectors with
!!  incompatible dimensions are added.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_add_DIMEN
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_add_DIMEN'
  TYPE(vectorT) :: test1, test2, test3
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: m=3
  INTEGER, PARAMETER :: expMsg=DIMEN
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize input vectors (data contents not important)
  CALL vec_init(test1, m)
  CALL vec_init(test2, m+1)  !> note different number of entries

  !> attempt to add the vectors
  test3 = vec_add(test1,test2, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate vectors
  CALL log_closeLogFile()
  CALL vec_clean(test1)
  CALL vec_clean(test2)
  CALL vec_clean(test3)

END SUBROUTINE test_vec_add_DIMEN


! ------------------------------------------------------------------------
!>  \test Test for correct addition of vectors
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      First input vector
!!  \param test2      Second input vector
!!  \param test3      Output vector
!!  \param m          Number of rows in the vector
!!  \param expDat     Expected result of addition
!!
!!  This test checks that the correct result is obtained when adding two
!!  dense vectors.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_add_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_add_VAL'
  TYPE(vectorT) :: test1, test2, test3
  INTEGER, PARAMETER :: m=3
  DOUBLE PRECISION, DIMENSION(m) :: expDat
  INTEGER :: i    !> loop variable

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize expected result
  expDat = RESHAPE(  (/ 3.d0,3.d0,3.d0 /), SHAPE(expDat) )

  !> initialize first input vector
  !> (loop sets vector to all ones)
  CALL vec_init(test1, m)
  DO i = 1,m
    CALL vec_set(test1, i, 1.d0)
  END DO

  !> initialize second input vector
  !> (loop sets vector to all twos)
  CALL vec_init(test2, m)
  DO i = 1,m
    CALL vec_set(test2, i, 2.d0)
  END DO

  !> add the vectors
  test3 = vec_add(test1,test2)
  CALL assertEquals(expDat,test3%dat, m)

  !> deallocate vectors
  CALL vec_clean(test1)
  CALL vec_clean(test2)
  CALL vec_clean(test3)

END SUBROUTINE test_vec_add_VAL


! ------------------------------------------------------------------------
!>  \test Test for correct addition of vectors ( using OPERATOR (+) )
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      First input vector
!!  \param test2      Second input vector
!!  \param test3      Output vector
!!  \param m          Number of rows in the vector
!!  \param expDat     Expected result of addition
!!
!!  This test checks that the correct result is obtained when adding two
!!  dense vectors. It uses the overloaded operator (+).
! ------------------------------------------------------------------------
SUBROUTINE test_vec_add_OP
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_add_OP'
  TYPE(vectorT) :: test1, test2, test3
  INTEGER, PARAMETER :: m=3
  DOUBLE PRECISION, DIMENSION(m) :: expDat
  INTEGER :: i    !> loop variable

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize expected result
  expDat = RESHAPE(  (/ 3.d0,3.d0,3.d0 /), SHAPE(expDat) )

  !> initialize first input vector
  !> (loop sets vector to all ones)
  CALL vec_init(test1, m)
  DO i = 1,m
    CALL vec_set(test1, i, 1.d0)
  END DO

  !> initialize second input vector
  !> (loop sets vector to all twos)
  CALL vec_init(test2, m)
  DO i = 1,m
    CALL vec_set(test2, i, 2.d0)
  END DO

  !> add the vectors
  test3 = test1 + test2
  CALL assertEquals(expDat,test3%dat, m)

  !> deallocate vectors
  CALL vec_clean(test1)
  CALL vec_clean(test2)
  CALL vec_clean(test3)

END SUBROUTINE test_vec_add_OP


! ------------------------------------------------------------------------
!>  \test Test for DIMEN exception when performing mapped add
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      Test target vector
!!  \param test2      Test input vector
!!  \param testName   Filename for log file (required for exceptions)
!!  \param m          Number of entries in target vector
!!  \param n          Number of entries in input vector
!!  \param expMsg     Expected exception message
!!  \param actMsg     Actual exception message
!!  \param ind        Mapping indices (size does not match test2)
!!
!!  This test checks that a DIMEN exception is raised when mapped addition
!!  is attempted with the dimension of the index array not matching the
!!  dimensions of the input vector.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_mapped_add_DIMEN
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_mapped_add_DIMEN'
  TYPE(vectorT) :: test1, test2
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: m=5,n=3
  INTEGER, PARAMETER :: expMsg=DIMEN
  INTEGER, PARAMETER, DIMENSION(2) :: ind = (/ 1,2 /)
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize vectors (data content not important)
  CALL vec_init(test1,m)
  CALL vec_init(test2,n)

  !> try mapped add (note n=3, SIZE(ind)=2)
  CALL vec_mappedAdd(test1, test2,ind, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate vectors
  CALL log_closeLogFile()
  CALL vec_clean(test1)
  CALL vec_clean(test2)

END SUBROUTINE test_vec_mapped_add_DIMEN


! ------------------------------------------------------------------------
!>  \test Test for POSIT exception when performing mapped add
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      Test target vector
!!  \param test2      Test input vector
!!  \param testName   Filename for log file (required for exceptions)
!!  \param m          Number of entries in target vector
!!  \param n          Number of entries in input vector
!!  \param expMsg     Expected exception message
!!  \param actMsg     Actual exception message
!!  \param ind1       Mapping indices (with index below minimum)
!!  \param ind2       Mapping indices (with index above maximum)
!!
!!  This test checks that a POSIT exception is raised when mapped addition
!!  is attempted with mapping indices that are outside the target vector.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_mapped_add_POSIT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_mapped_add_POSIT'
  TYPE(vectorT) :: test1, test2
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: m=5,n=3
  INTEGER, PARAMETER :: expMsg=POSIT
  INTEGER, PARAMETER, DIMENSION(n) :: ind1 = (/ -1,3,5 /)
  INTEGER, PARAMETER, DIMENSION(n) :: ind2 = (/ 1,3,6 /)
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize vectors (data not important)
  CALL vec_init(test1,m)
  CALL vec_init(test2,n)

  !> attempt first mapped add
  CALL vec_mappedAdd(test1, test2,ind1, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> attempt second mapped add
  CALL vec_mappedAdd(test1, test2,ind2, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate vectors
  CALL log_closeLogFile()
  CALL vec_clean(test1)
  CALL vec_clean(test2)

END SUBROUTINE test_vec_mapped_add_POSIT


! ------------------------------------------------------------------------
!>  \test Test for correct mapped addition (single operation)
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      Test target vector
!!  \param test2      Test input vector
!!  \param m          Number of entries in target vector
!!  \param n          Number of entries in input vector
!!  \param expDat     Expected data contents
!!  \param ind        Mapping indices
!!
!!  This test checks that mapped addition is performed correctly. The
!!  operation is performed once on a blank target vector.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_mapped_add_VAL1
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_mapped_add_VAL1'
  TYPE(vectorT) :: test1, test2
  INTEGER, PARAMETER :: m=5,n=3
  INTEGER, PARAMETER, DIMENSION(n) :: ind = (/ 1,3,5 /)
  DOUBLE PRECISION, DIMENSION(m) :: expDat
  INTEGER :: i    !> loop variable

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set up expected data
  expDat = RESHAPE( (/ 1.d0,0.d0,2.d0,0.d0,3.d0 /), SHAPE(expDat) )

  !> initialize target vector
  CALL vec_init(test1,m)

  !> initialize input vector
  !> (loop sets the data contents)
  CALL vec_init(test2,n)
  DO i = 1,n
    CALL vec_set(test2,i,DBLE(i))
  END DO

  !> perform mapped addition
  CALL vec_mappedAdd(test1, test2,ind)
  CALL assertEquals(expDat,test1%dat,m)

  !> deallocate vectors
  CALL vec_clean(test1)
  CALL vec_clean(test2)

END SUBROUTINE test_vec_mapped_add_VAL1


! ------------------------------------------------------------------------
!>  \test Test for correct mapped addition (zero index)
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      Test target vector
!!  \param test2      Test input vector
!!  \param m          Number of entries in target vector
!!  \param n          Number of entries in input vector
!!  \param expDat     Expected data contents
!!  \param ind        Mapping indices (with zero index indicating no-op)
!!
!!  This test checks that mapped addition is performed correctly. The
!!  mapping index contains a zero, which should tell the mapped adder to
!!  skip that location in the input vector.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_mapped_add_VAL2
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_mapped_add_VAL2'
  TYPE(vectorT) :: test1, test2
  INTEGER, PARAMETER :: m=5,n=3
  INTEGER, PARAMETER, DIMENSION(n) :: ind = (/ 1,0,5 /)
  DOUBLE PRECISION, DIMENSION(m) :: expDat
  INTEGER :: i    !> loop variable

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set up expected data
  expDat = RESHAPE( (/ 1.d0,0.d0,0.d0,0.d0,3.d0 /), SHAPE(expDat) )

  !> initialize target vector
  CALL vec_init(test1,m)

  !> initialize input vector
  !> (loop sets the data contents)
  CALL vec_init(test2,n)
  DO i = 1,n
    CALL vec_set(test2,i,DBLE(i))
  END DO

  !> perform mapped addition
  CALL vec_mappedAdd(test1, test2,ind)
  CALL assertEquals(expDat,test1%dat,m)

  !> deallocate vectors
  CALL vec_clean(test1)
  CALL vec_clean(test2)

END SUBROUTINE test_vec_mapped_add_VAL2


! ------------------------------------------------------------------------
!>  \test Test for correct mapped addition (multiple operations)
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      Test target vector
!!  \param test2      Test input vector
!!  \param m          Number of entries in target vector
!!  \param n          Number of entries in input vector
!!  \param expDat     Expected data contents
!!  \param ind2       Mapping indices for first add
!!  \param ind3       Mapping indices for second add
!!
!!  This test checks that mapped addition is performed correctly. The
!!  operation is performed twice to ensure summation of overlapping
!!  values. That is, input is added to existing data in the target vector
!!  rather than overwriting.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_mapped_add_VAL3
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_mapped_add_VAL3'
  TYPE(vectorT) :: test1, test2
  INTEGER, PARAMETER :: m=5,n=3
  INTEGER, PARAMETER, DIMENSION(n) :: ind1 = (/ 1,3,5 /)
  INTEGER, PARAMETER, DIMENSION(n) :: ind2 = (/ 1,0,5 /)
  DOUBLE PRECISION, DIMENSION(m) :: expDat
  INTEGER :: i    !> loop variable

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set up expected data
  expDat = RESHAPE( (/ 2.d0,0.d0,2.d0,0.d0,6.d0 /), SHAPE(expDat) )

  !> initialize target vector
  CALL vec_init(test1,m)

  !> initialize input vector
  !> (loop sets the data contents)
  CALL vec_init(test2,n)
  DO i = 1,n
    CALL vec_set(test2,i,DBLE(i))
  END DO

  !> perform the mapped addition for both input indices
  CALL vec_mappedAdd(test1, test2,ind1)
  CALL vec_mappedAdd(test1, test2,ind2)
  CALL assertEquals(expDat,test1%dat,m)

  !> deallocate vectors
  CALL vec_clean(test1)
  CALL vec_clean(test2)

END SUBROUTINE test_vec_mapped_add_VAL3


! ------------------------------------------------------------------------
!>  \test Test for scalar multiplication
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test1    Input vector
!!  \param test2    Output vector
!!  \param k        Scalar coefficient
!!  \param m        Number of entries in the vector
!!  \param expDat   Expected data contents
!!
!!  This test checks that scalar multiplication is performed correctly.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_scal_mul_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_scal_mul_VAL'
  TYPE(vectorT) :: test1, test2
  DOUBLE PRECISION, PARAMETER :: k=3.d0
  INTEGER, PARAMETER :: m=3
  DOUBLE PRECISION, DIMENSION(m) :: expDat
  INTEGER :: i    !> loop variable

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set up expected data
  expDat = RESHAPE( (/ 3.d0,3.d0,3.d0 /), SHAPE(expDat) )

  !> initialize input vector
  !> (loop sets the vector to ones)
  CALL vec_init(test1, m)
  DO i = 1,m
    CALL vec_set(test1, i, 1.d0)
  END DO

  !> multiply by scalar
  test2 = vec_scalMul(test1,k)
  CALL assertEquals(expDat,test2%dat, m)

  !> deallocate vectors
  CALL vec_clean(test1)
  CALL vec_clean(test2)

END SUBROUTINE test_vec_scal_mul_VAL


! ------------------------------------------------------------------------
!>  \test Test for scalar multiplication (by zero)
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test1    Input vector
!!  \param test2    Output vector
!!  \param k        Scalar coefficient
!!  \param m        Number of entries in the vector
!!  \param expDat   Expected data contents
!!
!!  This test checks that scalar multiplication is performed correctly
!!  when the scalar is zero.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_scal_mul_ZERO
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_scal_mul_ZERO'
  TYPE(vectorT) :: test1, test2
  DOUBLE PRECISION, PARAMETER :: k=0.d0
  INTEGER, PARAMETER :: m=3
  DOUBLE PRECISION, DIMENSION(m) :: expDat
  INTEGER :: i    !> loop variable

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set up expected data
  expDat = RESHAPE( (/ 0.d0,0.d0,0.d0 /), SHAPE(expDat) )

  !> initialize input vector
  !> (loop sets the vector to ones)
  CALL vec_init(test1, m)
  DO i = 1,m
    CALL vec_set(test1, i, 1.d0)
  END DO

  !> multiply by zero
  test2 = vec_scalMul(test1,k)
  CALL assertEquals(expDat,test2%dat, m)

  !> deallocate vectors
  CALL vec_clean(test1)
  CALL vec_clean(test2)

END SUBROUTINE test_vec_scal_mul_ZERO


! ------------------------------------------------------------------------
!>  \test Test for scalar multiplication ( using OPERATOR (*) )
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test1    Input vector
!!  \param test2    Output vector
!!  \param k        Scalar coefficient
!!  \param m        Number of entries in the vector
!!  \param expDat   Expected data contents
!!
!!  This test checks that scalar multiplication is performed correctly. It
!!  uses the overloaded operator (*).
! ------------------------------------------------------------------------
SUBROUTINE test_vec_scal_mul_OP
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_scal_mul_OP'
  TYPE(vectorT) :: test1, test2
  DOUBLE PRECISION, PARAMETER :: k=3.d0
  INTEGER, PARAMETER :: m=3
  DOUBLE PRECISION, DIMENSION(m) :: expDat
  INTEGER :: i    !> loop variable

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set up expected data
  expDat = RESHAPE( (/ 3.d0,3.d0,3.d0 /), SHAPE(expDat) )

  !> initialize input vector
  !> (loop sets the vector to ones)
  CALL vec_init(test1, m)
  DO i = 1,m
    CALL vec_set(test1, i, 1.d0)
  END DO

  !> multiply by scalar
  test2 = k*test1
  CALL assertEquals(expDat,test2%dat, m)

  !> reset output vector
  CALL vec_clean(test2)

  !> multiply by scalar (switch order of inputs)
  test2 = test1*k
  CALL assertEquals(expDat,test2%dat, m)

  !> deallocate vectors
  CALL vec_clean(test1)
  CALL vec_clean(test2)

END SUBROUTINE test_vec_scal_mul_OP


! ------------------------------------------------------------------------
!>  \test Test for DIMEN exception when performing dot product
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      First test vector
!!  \param test2      Second test vector
!!  \param testName   Filename for log file (required for exceptions)
!!  \param m          Number of entries in vector
!!  \param expMsg     Expected exception message
!!  \param actMsg     Actual exception message
!!  \param v          Dummy variable for dot product output
!!
!!  This test checks that a DIMEN exception is raised when a dot product
!!  is attempted when the vectors do not have the same length.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_dot_prod_DIMEN
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_dot_prod_DIMEN'
  TYPE(vectorT) :: test1, test2
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: m=3
  INTEGER, PARAMETER :: expMsg=DIMEN
  INTEGER :: actMsg
  DOUBLE PRECISION :: v

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> set up vectors (data contents not important)
  CALL vec_init(test1,m)
  CALL vec_init(test2,m+1)  !> note the different dimension

  !> attempt the dot product
  v = vec_dotProd(test1, test2, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate vectors
  CALL log_closeLogFile()
  CALL vec_clean(test1)
  CALL vec_clean(test2)

END SUBROUTINE test_vec_dot_prod_DIMEN


! ------------------------------------------------------------------------
!>  \test Test for dot product (with zero vector)
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test1    First input vector (zeros)
!!  \param test2    Second input vector
!!  \param m        Number of entries in the vector
!!  \param expected Expected result
!!  \param actual   Actual result
!!
!!  This test checks that the dot product result is zero when one vector
!!  is all zeros.
! ------------------------------------------------------------------------
SUBROUTINE test_vec_dot_prod_ZERO
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_dot_prod_ZERO'
  TYPE(vectorT) :: test1, test2
  INTEGER, PARAMETER :: m=3
  DOUBLE PRECISION, PARAMETER :: expected=0.d0
  DOUBLE PRECISION :: actual
  INTEGER :: i    !> loop variable

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize first input vector (zeros)
  CALL vec_init(test1,m)

  !> initialize second input vector
  !! (loop sets up data content)
  CALL vec_init(test2,m)
  DO i = 1,m
    CALL vec_set(test2, i, ( 2.d0 * DBLE(i) ) )
  END DO

  !> perform dot product
  actual = vec_dotProd(test1, test2)
  CALL assertEquals(expected,actual)

  !> deallocate vectors
  CALL vec_clean(test1)
  CALL vec_clean(test2)

END SUBROUTINE test_vec_dot_prod_ZERO


! ------------------------------------------------------------------------
!>  \test Test for correct dot product
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test1    First input vector
!!  \param test2    Second input vector
!!  \param m        Number of entries in the vector
!!  \param expected Expected result
!!  \param actual   Actual result
!!
!!  This test checks that the dot product result is correct in the general
!!  case (when the data content is non-zero).
! ------------------------------------------------------------------------
SUBROUTINE test_vec_dot_prod_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_vec_dot_prod_VAL'
  TYPE(vectorT) :: test1, test2
  INTEGER, PARAMETER :: m=3
  DOUBLE PRECISION, PARAMETER :: expected=28.d0
  DOUBLE PRECISION :: actual
  INTEGER :: i    !> loop variable

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize first input vector
  !! (loop sets up data content)
  CALL vec_init(test1,m)
  DO i = 1,m
    CALL vec_set(test1, i, DBLE(i))
  END DO

  !> initialize second input vector
  !! (loop sets up data content)
  CALL vec_init(test2,m)
  DO i = 1,m
    CALL vec_set(test2, i, ( 2.d0 * DBLE(i) ) )
  END DO

  !> perform dot product
  actual = vec_dotProd(test1, test2)
  CALL assertEquals(expected,actual)

  !> deallocate vectors
  CALL vec_clean(test1)
  CALL vec_clean(test2)

END SUBROUTINE test_vec_dot_prod_VAL

END MODULE vector_test
